from pytest_mock import MockerFixture


def test_server_error_counter():
    from apitally.client.server_errors import ServerErrorCounter

    server_errors = ServerErrorCounter()
    server_errors.add_server_error(
        consumer=None,
        method="GET",
        path="/test",
        exception=ValueError("test"),
    )
    server_errors.add_server_error(
        consumer=None,
        method="GET",
        path="/test",
        exception=ValueError("test"),
    )

    data = server_errors.get_and_reset_server_errors()
    assert len(server_errors.error_counts) == 0
    assert len(data) == 1
    assert data[0]["method"] == "GET"
    assert data[0]["path"] == "/test"
    assert data[0]["type"] == "builtins.ValueError"
    assert data[0]["msg"] == "test"
    assert data[0]["error_count"] == 2


def test_exception_truncation(mocker: MockerFixture):
    from apitally.client.server_errors import (
        get_exception_type,
        get_truncated_exception_msg,
        get_truncated_exception_traceback,
    )

    mocker.patch("apitally.client.server_errors.MAX_EXCEPTION_MSG_LENGTH", 32)
    mocker.patch("apitally.client.server_errors.MAX_EXCEPTION_TRACEBACK_LENGTH", 128)

    try:
        raise ValueError("a" * 88)
    except ValueError as e:
        type_ = get_exception_type(e)
        msg = get_truncated_exception_msg(e)
        tb = get_truncated_exception_traceback(e)

    assert type_ == "builtins.ValueError"
    assert len(msg) == 32
    assert msg.endswith("... (truncated)")
    assert len(tb) <= 128
    assert tb.startswith("... (truncated) ...\n")



def test_validation_error_counter():
    from apitally.client.validation_errors import ValidationErrorCounter

    validation_errors = ValidationErrorCounter()
    validation_errors.add_validation_errors(
        consumer=None,
        method="GET",
        path="/test",
        detail=[
            {
                "loc": ["query", "foo"],
                "type": "type_error.integer",
                "msg": "value is not a valid integer",
            },
            {
                "loc": ["query", "bar"],
                "type": "type_error.integer",
                "msg": "value is not a valid integer",
            },
        ],
    )
    validation_errors.add_validation_errors(
        consumer=None,
        method="GET",
        path="/test",
        detail=[
            {
                "loc": ["query", "foo"],
                "type": "type_error.integer",
                "msg": "value is not a valid integer",
            }
        ],
    )

    data = validation_errors.get_and_reset_validation_errors()
    assert len(validation_errors.error_counts) == 0
    assert len(data) == 2
    assert data[0]["method"] == "GET"
    assert data[0]["path"] == "/test"
    assert data[0]["loc"] == ("query", "foo")
    assert data[0]["type"] == "type_error.integer"
    assert data[0]["msg"] == "value is not a valid integer"
    assert data[0]["error_count"] == 2



from __future__ import annotations

import time
from contextlib import asynccontextmanager
from typing import Any, Awaitable, Callable, Dict, List, Optional, Union
from warnings import warn

from httpx import HTTPStatusError, Proxy
from starlette.applications import Starlette
from starlette.datastructures import Headers
from starlette.requests import Request
from starlette.routing import BaseRoute, Match, Router
from starlette.schemas import EndpointInfo, SchemaGenerator
from starlette.testclient import TestClient
from starlette.types import ASGIApp, Lifespan, Message, Receive, Scope, Send

from apitally.client.client_asyncio import ApitallyClient
from apitally.client.consumers import Consumer as ApitallyConsumer
from apitally.client.request_logging import (
    BODY_TOO_LARGE,
    MAX_BODY_SIZE,
    RequestLogger,
    RequestLoggingConfig,
)
from apitally.common import get_versions, parse_int, try_json_loads


__all__ = ["ApitallyMiddleware", "ApitallyConsumer", "RequestLoggingConfig"]


class ApitallyMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        client_id: str,
        env: str = "dev",
        request_logging_config: Optional[RequestLoggingConfig] = None,
        app_version: Optional[str] = None,
        openapi_url: Optional[str] = "/openapi.json",
        identify_consumer_callback: Optional[Callable[[Request], Union[str, ApitallyConsumer, None]]] = None,
        capture_client_disconnects: bool = False,
        proxy: Optional[Union[str, Proxy]] = None,
    ) -> None:
        self.app = app
        self.app_version = app_version
        self.openapi_url = openapi_url
        self.identify_consumer_callback = identify_consumer_callback
        self.capture_client_disconnects = capture_client_disconnects
        self.client = ApitallyClient(
            client_id=client_id,
            env=env,
            request_logging_config=request_logging_config,
            proxy=proxy,
        )

        self.capture_request_body = (
            self.client.request_logger.config.enabled and self.client.request_logger.config.log_request_body
        )
        self.capture_response_body = (
            self.client.request_logger.config.enabled and self.client.request_logger.config.log_response_body
        )

        _inject_lifespan_handlers(
            app,
            on_startup=self.on_startup,
            on_shutdown=self.client.handle_shutdown,
        )

    async def on_startup(self) -> None:
        data = _get_startup_data(self.app, app_version=self.app_version, openapi_url=self.openapi_url)
        self.client.set_startup_data(data)
        self.client.start_sync_loop()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if self.client.enabled and scope["type"] == "http" and scope["method"] != "OPTIONS":
            timestamp = time.time()
            request = Request(scope, receive, send)
            request_size = parse_int(request.headers.get("Content-Length"))
            request_body = b""
            request_body_too_large = request_size is not None and request_size > MAX_BODY_SIZE
            response_status = 0
            response_time: Optional[float] = None
            response_headers = Headers()
            response_body = b""
            response_body_too_large = False
            response_size: Optional[int] = None
            response_chunked = False
            response_content_type: Optional[str] = None
            exception: Optional[BaseException] = None
            start_time = time.perf_counter()

            async def receive_wrapper() -> Message:
                nonlocal request_body, request_body_too_large
                message = await receive()
                if message["type"] == "http.request" and self.capture_request_body and not request_body_too_large:
                    request_body += message.get("body", b"")
                    if len(request_body) > MAX_BODY_SIZE:
                        request_body_too_large = True
                        request_body = b""
                return message

            async def send_wrapper(message: Message) -> None:
                nonlocal \
                    response_time, \
                    response_status, \
                    response_headers, \
                    response_body, \
                    response_body_too_large, \
                    response_chunked, \
                    response_content_type, \
                    response_size
                if message["type"] == "http.response.start":
                    response_time = time.perf_counter() - start_time
                    response_status = message["status"]
                    response_headers = Headers(scope=message)
                    response_chunked = (
                        response_headers.get("Transfer-Encoding") == "chunked"
                        or "Content-Length" not in response_headers
                    )
                    response_content_type = response_headers.get("Content-Type")
                    response_size = parse_int(response_headers.get("Content-Length")) if not response_chunked else 0
                    response_body_too_large = response_size is not None and response_size > MAX_BODY_SIZE
                elif message["type"] == "http.response.body":
                    if response_chunked and response_size is not None:
                        response_size += len(message.get("body", b""))
                    if (
                        (self.capture_response_body or response_status == 422)
                        and RequestLogger.is_supported_content_type(response_content_type)
                        and not response_body_too_large
                    ):
                        response_body += message.get("body", b"")
                        if len(response_body) > MAX_BODY_SIZE:
                            response_body_too_large = True
                            response_body = b""
                await send(message)

            try:
                await self.app(scope, receive_wrapper, send_wrapper)
            except BaseException as e:
                exception = e
                raise e from None
            finally:
                if response_time is None:
                    response_time = time.perf_counter() - start_time
                if self.capture_client_disconnects and await request.is_disconnected():
                    # Client closed connection (report NGINX specific status code)
                    response_status = 499
                self.add_request(
                    timestamp=timestamp,
                    request=request,
                    request_body=request_body if not request_body_too_large else BODY_TOO_LARGE,
                    request_size=request_size,
                    response_status=response_status,
                    response_time=response_time,
                    response_headers=response_headers,
                    response_body=response_body if not response_body_too_large else BODY_TOO_LARGE,
                    response_size=response_size,
                    exception=exception,
                )
        else:
            await self.app(scope, receive, send)  # pragma: no cover

    def add_request(
        self,
        timestamp: float,
        request: Request,
        request_body: bytes,
        request_size: Optional[int],
        response_status: int,
        response_time: float,
        response_headers: Headers,
        response_body: bytes,
        response_size: Optional[int],
        exception: Optional[BaseException] = None,
    ) -> None:
        path = self.get_path(request)

        consumer = self.get_consumer(request)
        consumer_identifier = consumer.identifier if consumer else None
        self.client.consumer_registry.add_or_update_consumer(consumer)

        if path is not None:
            if response_status == 0 and exception is not None:
                response_status = 500
            self.client.request_counter.add_request(
                consumer=consumer_identifier,
                method=request.method,
                path=path,
                status_code=response_status,
                response_time=response_time,
                request_size=request_size,
                response_size=response_size,
            )
            if response_status == 422 and response_body and response_headers.get("Content-Type") == "application/json":
                body = try_json_loads(response_body, encoding=response_headers.get("Content-Encoding"))
                if isinstance(body, dict) and "detail" in body and isinstance(body["detail"], list):
                    # Log FastAPI / Pydantic validation errors
                    self.client.validation_error_counter.add_validation_errors(
                        consumer=consumer_identifier,
                        method=request.method,
                        path=path,
                        detail=body["detail"],
                    )
            if response_status == 500 and exception is not None:
                self.client.server_error_counter.add_server_error(
                    consumer=consumer_identifier,
                    method=request.method,
                    path=path,
                    exception=exception,
                )

        if self.client.request_logger.enabled:
            self.client.request_logger.log_request(
                request={
                    "timestamp": timestamp,
                    "method": request.method,
                    "path": path,
                    "url": str(request.url),
                    "headers": request.headers.items(),
                    "size": request_size,
                    "consumer": consumer_identifier,
                    "body": request_body,
                },
                response={
                    "status_code": response_status,
                    "response_time": response_time,
                    "headers": response_headers.items(),
                    "size": response_size,
                    "body": response_body,
                },
                exception=exception,
            )

    def get_path(self, request: Request, routes: Optional[list[BaseRoute]] = None) -> Optional[str]:
        if routes is None:
            routes = request.app.routes
        for route in routes:
            if hasattr(route, "routes"):
                path = self.get_path(request, routes=route.routes)
                if path is not None:
                    return path
            elif hasattr(route, "path"):
                match, _ = route.matches(request.scope)
                if match == Match.FULL:
                    return request.scope.get("root_path", "") + route.path
        return None

    def get_consumer(self, request: Request) -> Optional[ApitallyConsumer]:
        if hasattr(request.state, "apitally_consumer") and request.state.apitally_consumer:
            return ApitallyConsumer.from_string_or_object(request.state.apitally_consumer)
        if hasattr(request.state, "consumer_identifier") and request.state.consumer_identifier:
            # Keeping this for legacy support
            warn(
                "Providing a consumer identifier via `request.state.consumer_identifier` is deprecated, "
                "use `request.state.apitally_consumer` instead.",
                DeprecationWarning,
            )
            return ApitallyConsumer.from_string_or_object(request.state.consumer_identifier)
        if self.identify_consumer_callback is not None:
            consumer = self.identify_consumer_callback(request)
            return ApitallyConsumer.from_string_or_object(consumer)
        return None


def _get_startup_data(
    app: ASGIApp, app_version: Optional[str] = None, openapi_url: Optional[str] = None
) -> Dict[str, Any]:
    data: Dict[str, Any] = {}
    if openapi_url and (openapi := _get_openapi(app, openapi_url)):
        data["openapi"] = openapi
    if endpoints := _get_endpoint_info(app):
        data["paths"] = [{"path": endpoint.path, "method": endpoint.http_method} for endpoint in endpoints]
    data["versions"] = get_versions("fastapi", "starlette", app_version=app_version)
    data["client"] = "python:starlette"
    return data


def _get_openapi(app: ASGIApp, openapi_url: str) -> Optional[str]:
    try:
        client = TestClient(app, raise_server_exceptions=False)
        response = client.get(openapi_url)
        response.raise_for_status()
        return response.text
    except HTTPStatusError:
        return None


def _get_endpoint_info(app: ASGIApp) -> List[EndpointInfo]:
    routes = _get_routes(app)
    schemas = SchemaGenerator({})
    return schemas.get_endpoints(routes)


def _get_routes(app: Union[ASGIApp, Router]) -> List[BaseRoute]:
    if isinstance(app, Router):
        return app.routes
    elif hasattr(app, "app"):
        return _get_routes(app.app)
    return []  # pragma: no cover


def _inject_lifespan_handlers(
    app: Union[ASGIApp, Router],
    on_startup: Callable[[], Awaitable[Any]],
    on_shutdown: Callable[[], Awaitable[Any]],
) -> None:
    """
    Ensures the given startup and shutdown functions are called as part of the app's lifespan context manager.
    """
    router = app
    while not isinstance(router, Router) and hasattr(router, "app"):
        router = router.app
    if not isinstance(router, Router):
        raise TypeError("app must be a Starlette or Router instance")

    lifespan: Optional[Lifespan] = getattr(router, "lifespan_context", None)

    @asynccontextmanager
    async def wrapped_lifespan(app: Starlette):
        await on_startup()
        if lifespan is not None:
            async with lifespan(app):
                yield
        else:
            yield
        await on_shutdown()

    router.lifespan_context = wrapped_lifespan



def test_consumer_registry():
    from apitally.client.consumers import Consumer, ConsumerRegistry

    consumer_registry = ConsumerRegistry()
    consumer_registry.add_or_update_consumer(None)
    assert len(consumer_registry.consumers) == 0

    consumer_registry.add_or_update_consumer(Consumer("test"))
    assert len(consumer_registry.consumers) == 0

    consumer_registry.add_or_update_consumer(Consumer("test", name="Test"))
    consumer_registry.add_or_update_consumer(Consumer("test", group="Test"))
    assert len(consumer_registry.consumers) == 1
    assert consumer_registry.consumers["test"].name == "Test"
    assert consumer_registry.consumers["test"].group == "Test"

    consumer_registry.add_or_update_consumer(Consumer("test", group="Test 2"))
    assert len(consumer_registry.consumers) == 1
    assert consumer_registry.consumers["test"].name == "Test"
    assert consumer_registry.consumers["test"].group == "Test 2"

    consumer_registry.add_or_update_consumer(Consumer("test", name="Test 2"))
    assert len(consumer_registry.consumers) == 1
    assert consumer_registry.consumers["test"].name == "Test 2"
    assert consumer_registry.consumers["test"].group == "Test 2"

    assert len(consumer_registry.updated) == 1
    data = consumer_registry.get_and_reset_updated_consumers()
    assert len(data) == 1
    assert data[0]["identifier"] == "test"
    assert data[0]["name"] == "Test 2"
    assert data[0]["group"] == "Test 2"
    assert len(consumer_registry.updated) == 0
    assert len(consumer_registry.consumers) == 1



from __future__ import annotations

import asyncio
import logging
import random
import time
from contextlib import suppress
from functools import partial
from typing import Any, AsyncIterator, Dict, Optional, Union
from uuid import UUID

import backoff
import httpx

from apitally.client.client_base import MAX_QUEUE_TIME, REQUEST_TIMEOUT, ApitallyClientBase
from apitally.client.logging import get_logger
from apitally.client.request_logging import RequestLoggingConfig


logger = get_logger(__name__)
retry = partial(
    backoff.on_exception,
    backoff.expo,
    httpx.HTTPError,
    max_tries=3,
    logger=logger,
    giveup_log_level=logging.WARNING,
)


class ApitallyClient(ApitallyClientBase):
    def __init__(
        self,
        client_id: str,
        env: str,
        request_logging_config: Optional[RequestLoggingConfig] = None,
        proxy: Optional[Union[str, httpx.Proxy]] = None,
    ) -> None:
        super().__init__(client_id=client_id, env=env, request_logging_config=request_logging_config)
        self.proxy = proxy
        self._stop_sync_loop = False
        self._sync_loop_task: Optional[asyncio.Task] = None
        self._sync_data_queue: asyncio.Queue[Dict[str, Any]] = asyncio.Queue()

    def get_http_client(self) -> httpx.AsyncClient:
        if httpx.__version__ >= "0.26.0":
            # `proxy` parameter was added in version 0.26.0
            return httpx.AsyncClient(base_url=self.hub_url, timeout=REQUEST_TIMEOUT, proxy=self.proxy)
        else:
            return httpx.AsyncClient(base_url=self.hub_url, timeout=REQUEST_TIMEOUT, proxies=self.proxy)

    def start_sync_loop(self) -> None:
        self._stop_sync_loop = False
        self._sync_loop_task = asyncio.create_task(self._run_sync_loop())

    async def _run_sync_loop(self) -> None:
        last_sync_time = 0.0
        while not self._stop_sync_loop:
            try:
                self.request_logger.write_to_file()
            except Exception:  # pragma: no cover
                logger.exception("An error occurred while writing request logs")

            now = time.time()
            if (now - last_sync_time) >= self.sync_interval:
                try:
                    async with self.get_http_client() as client:
                        tasks = [self.send_sync_data(client), self.send_log_data(client)]
                        if not self._startup_data_sent:
                            tasks.append(self.send_startup_data(client))
                        await asyncio.gather(*tasks)
                    last_sync_time = now
                except Exception:  # pragma: no cover
                    logger.exception("An error occurred during sync with Apitally hub")

            self.request_logger.maintain()
            await asyncio.sleep(1)

    def stop_sync_loop(self) -> None:
        self._stop_sync_loop = True

    async def handle_shutdown(self) -> None:
        self.enabled = False
        if self._sync_loop_task is not None:
            self._sync_loop_task.cancel()
        # Send any remaining data before exiting
        async with self.get_http_client() as client:
            await self.send_sync_data(client)
            await self.send_log_data(client)

    def set_startup_data(self, data: Dict[str, Any]) -> None:
        self._startup_data_sent = False
        self._startup_data = self.add_uuids_to_data(data)

    async def send_startup_data(self, client: httpx.AsyncClient) -> None:
        if self._startup_data is not None:
            await self._send_startup_data(client, self._startup_data)

    async def send_sync_data(self, client: httpx.AsyncClient) -> None:
        data = self.get_sync_data()
        self._sync_data_queue.put_nowait(data)

        i = 0
        while not self._sync_data_queue.empty():
            data = self._sync_data_queue.get_nowait()
            try:
                if time.time() - data["timestamp"] <= MAX_QUEUE_TIME:
                    if i > 0:
                        await asyncio.sleep(random.uniform(0.1, 0.5))
                    await self._send_sync_data(client, data)
                    i += 1
            except httpx.HTTPError:
                self._sync_data_queue.put_nowait(data)
                break
            finally:
                self._sync_data_queue.task_done()

    async def send_log_data(self, client: httpx.AsyncClient) -> None:
        self.request_logger.rotate_file()
        i = 0
        while log_file := self.request_logger.get_file():
            if i > 0:
                time.sleep(random.uniform(0.1, 0.3))
            try:
                stream = log_file.stream_lines_compressed()
                await self._send_log_data(client, log_file.uuid, stream)
                log_file.delete()
            except httpx.HTTPError:
                self.request_logger.retry_file_later(log_file)
                break
            i += 1
            if i >= 10:
                break

    @retry(raise_on_giveup=False)
    async def _send_startup_data(self, client: httpx.AsyncClient, data: Dict[str, Any]) -> None:
        logger.debug("Sending startup data to Apitally hub")
        response = await client.post(url="/startup", json=data)
        self._handle_hub_response(response)
        self._startup_data_sent = True
        self._startup_data = None

    @retry()
    async def _send_sync_data(self, client: httpx.AsyncClient, data: Dict[str, Any]) -> None:
        logger.debug("Synchronizing data with Apitally hub")
        response = await client.post(url="/sync", json=data)
        self._handle_hub_response(response)

    async def _send_log_data(self, client: httpx.AsyncClient, uuid: UUID, stream: AsyncIterator[bytes]) -> None:
        logger.debug("Streaming request log data to Apitally hub")
        response = await client.post(url=f"{self.hub_url}/log?uuid={uuid}", content=stream)
        if response.status_code == 402 and "Retry-After" in response.headers:
            with suppress(ValueError):
                retry_after = int(response.headers["Retry-After"])
                self.request_logger.suspend_until = time.time() + retry_after
                self.request_logger.clear()
                return
        self._handle_hub_response(response)

    def _handle_hub_response(self, response: httpx.Response) -> None:
        if response.status_code == 404:
            self.enabled = False
            self.stop_sync_loop()
            logger.error("Invalid Apitally client ID: %s", self.client_id)
        elif response.status_code == 422:
            logger.error("Received validation error from Apitally hub: %s", response.json())
        else:
            response.raise_for_status()



from __future__ import annotations

import asyncio
import os
from asyncio import AbstractEventLoop
from typing import Iterator

import pytest


if os.getenv("PYTEST_RAISE", "0") != "0":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value


@pytest.fixture(scope="module")
def event_loop() -> Iterator[AbstractEventLoop]:
    policy = asyncio.get_event_loop_policy()
    loop = policy.new_event_loop()
    yield loop
    loop.close()



from typing import Dict

from django.http import HttpRequest
from django.urls import path
from ninja import NinjaAPI, Schema


api = NinjaAPI()


@api.get("/foo", summary="Foo", description="Foo")
def foo(request: HttpRequest) -> str:
    return "foo"


@api.get("/foo/{bar}")
def foo_bar(request: HttpRequest, bar: int) -> Dict[str, int]:
    return {"foo": bar}


class BarRequestBody(Schema):
    foo: str


@api.post("/bar")
def bar(request: HttpRequest, item: BarRequestBody) -> Dict[str, str]:
    return {"bar": item.foo}


@api.put("/baz")
def baz(request: HttpRequest) -> str:
    request.apitally_consumer = "baz"  # type: ignore[attr-defined]
    raise ValueError("baz")


@api.get("/val")
def val(request: HttpRequest, foo: int) -> str:
    return "val"


urlpatterns = [
    path("api/", api.urls),  # type: ignore[arg-type]
]



import json
import time
from typing import Callable, Dict, List, Optional, Union
from warnings import warn

from httpx import Proxy
from litestar.app import DEFAULT_OPENAPI_CONFIG, Litestar
from litestar.config.app import AppConfig
from litestar.connection import Request
from litestar.datastructures import Headers
from litestar.enums import ScopeType
from litestar.handlers import HTTPRouteHandler
from litestar.plugins import InitPluginProtocol
from litestar.types import ASGIApp, Message, Receive, Scope, Send

from apitally.client.client_asyncio import ApitallyClient
from apitally.client.consumers import Consumer as ApitallyConsumer
from apitally.client.request_logging import (
    BODY_TOO_LARGE,
    MAX_BODY_SIZE,
    RequestLogger,
    RequestLoggingConfig,
)
from apitally.common import get_versions, parse_int, try_json_loads


__all__ = ["ApitallyPlugin", "ApitallyConsumer", "RequestLoggingConfig"]


class ApitallyPlugin(InitPluginProtocol):
    def __init__(
        self,
        client_id: str,
        env: str = "dev",
        request_logging_config: Optional[RequestLoggingConfig] = None,
        app_version: Optional[str] = None,
        filter_openapi_paths: bool = True,
        identify_consumer_callback: Optional[Callable[[Request], Union[str, ApitallyConsumer, None]]] = None,
        proxy: Optional[Union[str, Proxy]] = None,
    ) -> None:
        self.client = ApitallyClient(
            client_id=client_id,
            env=env,
            request_logging_config=request_logging_config,
            proxy=proxy,
        )
        self.app_version = app_version
        self.filter_openapi_paths = filter_openapi_paths
        self.identify_consumer_callback = identify_consumer_callback

        self.openapi_path = "/schema"
        self.capture_request_body = (
            self.client.request_logger.config.enabled and self.client.request_logger.config.log_request_body
        )
        self.capture_response_body = (
            self.client.request_logger.config.enabled and self.client.request_logger.config.log_response_body
        )

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.on_startup.append(self.on_startup)
        app_config.on_shutdown.append(self.client.handle_shutdown)
        app_config.middleware.append(self.middleware_factory)
        app_config.after_exception.append(self.after_exception)
        return app_config

    def on_startup(self, app: Litestar) -> None:
        openapi_config = app.openapi_config or DEFAULT_OPENAPI_CONFIG
        if openapi_config.openapi_controller is not None:
            self.openapi_path = openapi_config.openapi_controller.path
        elif hasattr(openapi_config, "openapi_router") and openapi_config.openapi_router is not None:
            self.openapi_path = openapi_config.openapi_router.path
        elif openapi_config.path is not None:
            self.openapi_path = openapi_config.path

        data = {
            "openapi": _get_openapi(app),
            "paths": [route for route in _get_routes(app) if not self.filter_path(route["path"])],
            "versions": get_versions("litestar", app_version=self.app_version),
            "client": "python:litestar",
        }
        self.client.set_startup_data(data)
        self.client.start_sync_loop()

    def after_exception(self, exception: Exception, scope: Scope) -> None:
        scope["state"]["exception"] = exception

    def middleware_factory(self, app: ASGIApp) -> ASGIApp:
        async def middleware(scope: Scope, receive: Receive, send: Send) -> None:
            if self.client.enabled and scope["type"] == "http" and scope["method"] != "OPTIONS":
                timestamp = time.time()
                request = Request(scope)
                request_size = parse_int(request.headers.get("Content-Length"))
                request_body = b""
                request_body_too_large = request_size is not None and request_size > MAX_BODY_SIZE
                response_status = 0
                response_time = 0.0
                response_headers = Headers()
                response_body = b""
                response_body_too_large = False
                response_size: Optional[int] = None
                response_chunked = False
                response_content_type: Optional[str] = None
                start_time = time.perf_counter()

                async def receive_wrapper() -> Message:
                    nonlocal request_body, request_body_too_large
                    message = await receive()
                    if message["type"] == "http.request" and self.capture_request_body and not request_body_too_large:
                        request_body += message.get("body", b"")
                        if len(request_body) > MAX_BODY_SIZE:
                            request_body_too_large = True
                            request_body = b""
                    return message

                async def send_wrapper(message: Message) -> None:
                    nonlocal \
                        response_time, \
                        response_status, \
                        response_headers, \
                        response_body, \
                        response_body_too_large, \
                        response_chunked, \
                        response_content_type, \
                        response_size
                    if message["type"] == "http.response.start":
                        response_time = time.perf_counter() - start_time
                        response_status = message["status"]
                        response_headers = Headers(message["headers"])
                        response_chunked = (
                            response_headers.get("Transfer-Encoding") == "chunked"
                            or "Content-Length" not in response_headers
                        )
                        response_content_type = response_headers.get("Content-Type")
                        response_size = parse_int(response_headers.get("Content-Length")) if not response_chunked else 0
                        response_body_too_large = response_size is not None and response_size > MAX_BODY_SIZE
                    elif message["type"] == "http.response.body":
                        if response_chunked and response_size is not None:
                            response_size += len(message.get("body", b""))
                        if (
                            (self.capture_response_body or response_status == 400)
                            and RequestLogger.is_supported_content_type(response_content_type)
                            and not response_body_too_large
                        ):
                            response_body += message.get("body", b"")
                            if len(response_body) > MAX_BODY_SIZE:
                                response_body_too_large = True
                                response_body = b""
                    await send(message)

                await app(scope, receive_wrapper, send_wrapper)
                self.add_request(
                    timestamp=timestamp,
                    request=request,
                    request_body=request_body if not request_body_too_large else BODY_TOO_LARGE,
                    request_size=request_size,
                    response_status=response_status,
                    response_time=response_time,
                    response_headers=response_headers,
                    response_body=response_body if not response_body_too_large else BODY_TOO_LARGE,
                    response_size=response_size,
                )
            else:
                await app(scope, receive, send)  # pragma: no cover

        return middleware

    def add_request(
        self,
        timestamp: float,
        request: Request,
        request_body: bytes,
        request_size: Optional[int],
        response_status: int,
        response_time: float,
        response_headers: Headers,
        response_body: bytes,
        response_size: Optional[int],
    ) -> None:
        if response_status < 100:
            return  # pragma: no cover
        path = self.get_path(request)
        if self.filter_path(path):
            return

        consumer = self.get_consumer(request)
        consumer_identifier = consumer.identifier if consumer else None
        self.client.consumer_registry.add_or_update_consumer(consumer)

        if path is not None:
            self.client.request_counter.add_request(
                consumer=consumer_identifier,
                method=request.method,
                path=path,
                status_code=response_status,
                response_time=response_time,
                request_size=request_size,
                response_size=response_size,
            )

            if response_status == 400 and response_body and len(response_body) < 4096:
                body = try_json_loads(response_body, encoding=response_headers.get("Content-Encoding"))
                if (
                    isinstance(body, dict)
                    and "detail" in body
                    and isinstance(body["detail"], str)
                    and "validation" in body["detail"].lower()
                    and "extra" in body
                    and isinstance(body["extra"], list)
                ):
                    self.client.validation_error_counter.add_validation_errors(
                        consumer=consumer_identifier,
                        method=request.method,
                        path=path,
                        detail=[
                            {
                                "loc": [error.get("source", "body")] + error["key"].split("."),
                                "msg": error["message"],
                                "type": "",
                            }
                            for error in body["extra"]
                            if "key" in error and "message" in error
                        ],
                    )

            if response_status == 500 and "exception" in request.state:
                self.client.server_error_counter.add_server_error(
                    consumer=consumer_identifier,
                    method=request.method,
                    path=path,
                    exception=request.state["exception"],
                )

        if self.client.request_logger.enabled:
            self.client.request_logger.log_request(
                request={
                    "timestamp": timestamp,
                    "method": request.method,
                    "path": path,
                    "url": str(request.url),
                    "headers": [(k, v) for k, v in request.headers.items()],
                    "size": request_size,
                    "consumer": consumer_identifier,
                    "body": request_body,
                },
                response={
                    "status_code": response_status,
                    "response_time": response_time,
                    "headers": [(k, v) for k, v in response_headers.items()],
                    "size": response_size,
                    "body": response_body,
                },
                exception=request.state["exception"] if "exception" in request.state else None,
            )

    def get_path(self, request: Request) -> Optional[str]:
        if not request.route_handler.paths:
            return None
        path: List[str] = []
        for layer in request.route_handler.ownership_layers:
            if isinstance(layer, HTTPRouteHandler):
                if len(layer.paths) == 0:
                    return None  # pragma: no cover
                path.append(list(layer.paths)[0].lstrip("/"))
            else:
                path.append(layer.path.lstrip("/"))
        return "/" + "/".join(filter(None, path))

    def filter_path(self, path: Optional[str]) -> bool:
        if path is not None and self.filter_openapi_paths and self.openapi_path:
            return path == self.openapi_path or path.startswith(self.openapi_path + "/")
        return False  # pragma: no cover

    def get_consumer(self, request: Request) -> Optional[ApitallyConsumer]:
        if hasattr(request.state, "apitally_consumer") and request.state.apitally_consumer:
            return ApitallyConsumer.from_string_or_object(request.state.apitally_consumer)
        if hasattr(request.state, "consumer_identifier") and request.state.consumer_identifier:
            # Keeping this for legacy support
            warn(
                "Providing a consumer identifier via `request.state.consumer_identifier` is deprecated, "
                "use `request.state.apitally_consumer` instead.",
                DeprecationWarning,
            )
            return ApitallyConsumer.from_string_or_object(request.state.consumer_identifier)
        if self.identify_consumer_callback is not None:
            consumer = self.identify_consumer_callback(request)
            return ApitallyConsumer.from_string_or_object(consumer)
        return None


def _get_openapi(app: Litestar) -> str:
    schema = app.openapi_schema.to_schema()
    return json.dumps(schema)


def _get_routes(app: Litestar) -> List[Dict[str, str]]:
    return [
        {"method": method, "path": route.path}
        for route in app.routes
        for method in route.methods
        if route.scope_type == ScopeType.HTTP and method != "OPTIONS"
    ]



from __future__ import annotations

import json
import sys
from importlib.util import find_spec
from typing import TYPE_CHECKING, Optional

import pytest
from pytest_mock import MockerFixture


if find_spec("ninja") is None:
    pytest.skip("django-ninja is not available", allow_module_level=True)

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django.test import Client


def identify_consumer(request: HttpRequest) -> Optional[str]:
    return "test"


@pytest.fixture(scope="module")
def reset_modules() -> None:
    for module in list(sys.modules):
        if module.startswith("django.") or module.startswith("apitally."):
            del sys.modules[module]


@pytest.fixture(scope="module", autouse=True)
def setup(reset_modules, module_mocker: MockerFixture) -> None:
    import django
    from django.conf import settings

    from apitally.django_rest_framework import RequestLoggingConfig

    module_mocker.patch("apitally.client.client_threading.ApitallyClient._instance", None)
    module_mocker.patch("apitally.client.client_threading.ApitallyClient.start_sync_loop")
    module_mocker.patch("apitally.client.client_threading.ApitallyClient.set_startup_data")
    module_mocker.patch("apitally.django.ApitallyMiddleware.config", None)

    settings.configure(
        ROOT_URLCONF="tests.django_ninja_urls",
        ALLOWED_HOSTS=["testserver"],
        SECRET_KEY="secret",
        MIDDLEWARE=[
            "apitally.django_ninja.ApitallyMiddleware",
            "django.middleware.common.CommonMiddleware",
        ],
        APITALLY_MIDDLEWARE={
            "client_id": "76b5cb91-a0a4-4ea0-a894-57d2b9fcb2c9",
            "env": "dev",
            "request_logging_config": RequestLoggingConfig(
                enabled=True,
                log_request_body=True,
                log_response_body=True,
            ),
            "identify_consumer_callback": "tests.test_django_ninja.identify_consumer",
        },
    )
    django.setup()


@pytest.fixture(scope="module")
def client(module_mocker: MockerFixture) -> Client:
    import django
    from django.test import Client

    if django.VERSION[0] < 3:
        module_mocker.patch("django.test.client.Client.store_exc_info")  # Simulate raise_request_exception=False
    return Client(raise_request_exception=False)


def test_middleware_requests_ok(client: Client, mocker: MockerFixture):
    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")

    response = client.get("/api/foo/123")
    assert response.status_code == 200
    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["consumer"] == "test"
    assert mock.call_args.kwargs["method"] == "GET"
    assert mock.call_args.kwargs["path"] == "/api/foo/{bar}"
    assert mock.call_args.kwargs["status_code"] == 200
    assert mock.call_args.kwargs["response_time"] > 0
    assert int(mock.call_args.kwargs["response_size"]) > 0

    response = client.post("/api/bar", data={"foo": "bar"}, content_type="application/json")
    assert response.status_code == 200
    assert mock.call_count == 2
    assert mock.call_args is not None
    assert mock.call_args.kwargs["method"] == "POST"
    assert int(mock.call_args.kwargs["request_size"]) > 0


def test_middleware_requests_404(client: Client, mocker: MockerFixture):
    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")

    response = client.get("/api/none")
    assert response.status_code == 404
    mock.assert_not_called()


def test_middleware_requests_error(client: Client, mocker: MockerFixture):
    mock1 = mocker.patch("apitally.client.requests.RequestCounter.add_request")
    mock2 = mocker.patch("apitally.client.server_errors.ServerErrorCounter.add_server_error")

    response = client.put("/api/baz")
    assert response.status_code == 500
    mock1.assert_called_once()
    assert mock1.call_args is not None
    assert mock1.call_args.kwargs["method"] == "PUT"
    assert mock1.call_args.kwargs["path"] == "/api/baz"
    assert mock1.call_args.kwargs["status_code"] == 500
    assert mock1.call_args.kwargs["response_time"] > 0

    mock2.assert_called_once()
    assert mock2.call_args is not None
    exception = mock2.call_args.kwargs["exception"]
    assert isinstance(exception, ValueError)


def test_middleware_validation_error(client: Client, mocker: MockerFixture):
    mock = mocker.patch("apitally.client.validation_errors.ValidationErrorCounter.add_validation_errors")

    response = client.get("/api/val?foo=bar")
    assert response.status_code == 422
    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["method"] == "GET"
    assert mock.call_args.kwargs["path"] == "/api/val"
    assert len(mock.call_args.kwargs["detail"]) == 1
    assert mock.call_args.kwargs["detail"][0]["loc"] == ["query", "foo"]


def test_middleware_request_logging(client: Client, mocker: MockerFixture):
    from apitally.client.request_logging import BODY_TOO_LARGE

    mock = mocker.patch("apitally.client.request_logging.RequestLogger.log_request")

    response = client.get("/api/foo/123?foo=bar", HTTP_TEST_HEADER="test")
    assert response.status_code == 200
    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["method"] == "GET"
    assert mock.call_args.kwargs["request"]["path"] == "/api/foo/{bar}"
    assert mock.call_args.kwargs["request"]["url"] == "http://testserver/api/foo/123?foo=bar"
    assert ("Test-Header", "test") in mock.call_args.kwargs["request"]["headers"]
    assert mock.call_args.kwargs["request"]["consumer"] == "test"
    assert mock.call_args.kwargs["response"]["status_code"] == 200
    assert mock.call_args.kwargs["response"]["response_time"] > 0
    assert ("Content-Type", "application/json; charset=utf-8") in mock.call_args.kwargs["response"]["headers"]
    assert mock.call_args.kwargs["response"]["size"] > 0

    response = client.post("/api/bar", data={"foo": "foo"}, content_type="application/json")
    assert response.status_code == 200
    assert mock.call_count == 2
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["method"] == "POST"
    assert mock.call_args.kwargs["request"]["path"] == "/api/bar"
    assert mock.call_args.kwargs["request"]["url"] == "http://testserver/api/bar"
    assert mock.call_args.kwargs["request"]["body"] == b'{"foo": "foo"}'
    assert mock.call_args.kwargs["response"]["body"] == b'{"bar": "foo"}'

    mocker.patch("apitally.django.MAX_BODY_SIZE", 2)
    response = client.post("/api/bar", data={"foo": "foo"}, content_type="application/json")
    assert response.status_code == 200
    assert mock.call_count == 3
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["body"] == BODY_TOO_LARGE
    assert mock.call_args.kwargs["response"]["body"] == BODY_TOO_LARGE


def test_get_startup_data():
    from apitally.django import _get_startup_data

    data = _get_startup_data(app_version="1.2.3", urlconfs=[None])
    openapi = json.loads(data["openapi"])
    assert len(data["paths"]) == 5
    assert len(openapi["paths"]) == 5

    assert data["versions"]["django"]
    assert data["versions"]["django-ninja"]
    assert data["versions"]["app"] == "1.2.3"
    assert data["client"] == "python:django"


def test_get_ninja_api_instances():
    from ninja import NinjaAPI

    from apitally.django import _get_ninja_api_instances

    apis = _get_ninja_api_instances()
    assert len(apis) == 1
    api = list(apis)[0]
    assert isinstance(api, NinjaAPI)


def test_get_ninja_api_endpoints():
    from apitally.django import _get_ninja_paths

    endpoints = _get_ninja_paths([None])
    assert len(endpoints) == 5
    assert all(len(e["summary"]) > 0 for e in endpoints)
    assert any(e["description"] is not None and len(e["description"]) > 0 for e in endpoints)


def test_check_import():
    from apitally.django import _check_import

    assert _check_import("ninja") is True
    assert _check_import("nonexistentpackage") is False



from __future__ import annotations

import asyncio
import gzip
import json
import re
import time
from typing import TYPE_CHECKING

import pytest
import pytest_httpx
from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture

from .constants import CLIENT_ID, ENV


if TYPE_CHECKING:
    from apitally.client.client_asyncio import ApitallyClient


@pytest.fixture(scope="module")
async def client() -> ApitallyClient:
    from apitally.client.client_asyncio import ApitallyClient, RequestLoggingConfig

    client = ApitallyClient(
        client_id=CLIENT_ID,
        env=ENV,
        request_logging_config=RequestLoggingConfig(enabled=True),
    )
    client.request_counter.add_request(
        consumer=None,
        method="GET",
        path="/test",
        status_code=200,
        response_time=0.105,
    )
    client.request_counter.add_request(
        consumer=None,
        method="GET",
        path="/test",
        status_code=200,
        response_time=0.227,
    )
    client.request_counter.add_request(
        consumer=None,
        method="GET",
        path="/test",
        status_code=422,
        response_time=0.02,
    )
    client.validation_error_counter.add_validation_errors(
        consumer=None,
        method="GET",
        path="/test",
        detail=[
            {
                "loc": ["query", "foo"],
                "type": "type_error.integer",
                "msg": "value is not a valid integer",
            },
        ],
    )
    return client


def log_request(client: ApitallyClient) -> None:
    client.request_logger.log_request(
        request={
            "timestamp": time.time(),
            "method": "GET",
            "path": "/test",
            "url": "http://testserver/test",
            "headers": [],
            "size": 0,
            "consumer": None,
            "body": None,
        },
        response={
            "status_code": 200,
            "response_time": 0.105,
            "headers": [],
            "size": 0,
            "body": None,
        },
    )
    client.request_logger.write_to_file()


async def test_sync_loop(client: ApitallyClient, mocker: MockerFixture):
    send_sync_data_mock = mocker.patch("apitally.client.client_asyncio.ApitallyClient.send_sync_data")
    mocker.patch("apitally.client.client_base.INITIAL_SYNC_INTERVAL", 0.05)

    client.start_sync_loop()
    await asyncio.sleep(0.2)  # Ensure loop starts
    client.stop_sync_loop()  # Should stop after next iteration
    await asyncio.sleep(0.1)  # Wait for task to finish
    assert send_sync_data_mock.await_count >= 1


async def test_send_sync_data(client: ApitallyClient, httpx_mock: HTTPXMock):
    from apitally.client.client_base import HUB_BASE_URL, HUB_VERSION

    httpx_mock.add_response()
    async with client.get_http_client() as http_client:
        await client.send_sync_data(client=http_client)

    request = httpx_mock.get_request(url=f"{HUB_BASE_URL}/{HUB_VERSION}/{CLIENT_ID}/{ENV}/sync")
    assert request is not None
    request_data = json.loads(request.read())
    assert len(request_data["requests"]) == 2
    assert request_data["requests"][0]["request_count"] == 2
    assert len(request_data["validation_errors"]) == 1
    assert request_data["validation_errors"][0]["error_count"] == 1


async def test_send_log_data(client: ApitallyClient, httpx_mock: HTTPXMock):
    from apitally.client.client_base import HUB_BASE_URL, HUB_VERSION

    log_request(client)
    httpx_mock.add_response()
    async with client.get_http_client() as http_client:
        await client.send_log_data(client=http_client)

    url_pattern = re.compile(rf"{HUB_BASE_URL}/{HUB_VERSION}/{CLIENT_ID}/{ENV}/log\?uuid=[a-f0-9-]+$")
    request = httpx_mock.get_request(url=url_pattern)
    assert request is not None

    json_lines = gzip.decompress(request.read()).strip().split(b"\n")
    assert len(json_lines) == 1
    json_data = json.loads(json_lines[0])
    assert json_data["request"]["path"] == "/test"
    assert json_data["response"]["status_code"] == 200

    if pytest_httpx.__version__ < "0.31.0":
        httpx_mock.reset(True)  # type: ignore[call-arg]
    else:
        httpx_mock.reset()

    # Test 402 response with Retry-After header
    log_request(client)
    httpx_mock.add_response(status_code=402, headers={"Retry-After": "3600"})
    async with client.get_http_client() as http_client:
        await client.send_log_data(client=http_client)
    assert httpx_mock.get_request(url=url_pattern) is not None
    assert client.request_logger.suspend_until is not None
    assert client.request_logger.suspend_until > time.time() + 3590

    # Ensure not logging requests anymore
    log_request(client)
    assert client.request_logger.file is None
    assert len(client.request_logger.write_deque) == 0
    assert len(client.request_logger.file_deque) == 0


async def test_send_startup_data(client: ApitallyClient, httpx_mock: HTTPXMock):
    from apitally.client.client_base import HUB_BASE_URL, HUB_VERSION

    httpx_mock.add_response()
    data = {"paths": [], "client_version": "1.0.0", "starlette_version": "0.28.0", "python_version": "3.11.4"}
    client.set_startup_data(data)
    async with client.get_http_client() as http_client:
        await client.send_startup_data(client=http_client)

    request = httpx_mock.get_request(url=f"{HUB_BASE_URL}/{HUB_VERSION}/{CLIENT_ID}/{ENV}/startup")
    assert request is not None
    request_data = json.loads(request.read())
    assert request_data["paths"] == []
    assert request_data["client_version"] == "1.0.0"



from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING, Optional

import pytest
from pytest_mock import MockerFixture

from .constants import CLIENT_ID, ENV


if find_spec("litestar") is None:
    pytest.skip("litestar is not available", allow_module_level=True)
else:
    # Need to import Stream at package level to avoid NameError in Litestar
    from litestar.response import Stream


if TYPE_CHECKING:
    from litestar.app import Litestar
    from litestar.testing import TestClient


@pytest.fixture(scope="module")
async def app(module_mocker: MockerFixture) -> Litestar:
    from litestar.app import Litestar
    from litestar.connection import Request
    from litestar.handlers import get, post
    from litestar.response import Stream

    from apitally.litestar import ApitallyConsumer, ApitallyPlugin, RequestLoggingConfig

    async def mocked_handle_shutdown(_):
        # Empty function instead of Mock to avoid the following error in Python 3.10:
        # TypeError: 'Mock' object is not subscriptable
        pass

    module_mocker.patch("apitally.client.client_asyncio.ApitallyClient._instance", None)
    module_mocker.patch("apitally.client.client_asyncio.ApitallyClient.start_sync_loop")
    module_mocker.patch("apitally.client.client_asyncio.ApitallyClient.handle_shutdown", mocked_handle_shutdown)

    @get("/foo")
    async def foo() -> str:
        return "foo"

    @get("/foo/{bar:str}")
    async def foo_bar(request: Request, bar: str) -> str:
        request.state.apitally_consumer = "test2"
        return f"foo: {bar}"

    @post("/bar")
    async def bar(request: Request) -> str:
        body = await request.body()
        return "bar: " + body.decode()

    @post("/baz")
    async def baz() -> None:
        raise ValueError("baz")

    @get("/val")
    async def val(foo: int) -> str:
        return "val"

    @get("/stream")
    async def stream() -> Stream:
        def stream_response():
            yield b"foo"
            yield b"bar"

        return Stream(stream_response())

    def identify_consumer(request: Request) -> Optional[ApitallyConsumer]:
        return ApitallyConsumer("test1", name="Test 1") if "/foo" in request.route_handler.paths else None

    plugin = ApitallyPlugin(
        client_id=CLIENT_ID,
        env=ENV,
        app_version="1.2.3",
        identify_consumer_callback=identify_consumer,
        request_logging_config=RequestLoggingConfig(
            enabled=True,
            log_request_body=True,
            log_response_body=True,
        ),
    )
    app = Litestar(
        route_handlers=[foo, foo_bar, bar, baz, val, stream],
        plugins=[plugin],
    )
    return app


@pytest.fixture(scope="module")
async def client(app: Litestar) -> TestClient:
    from litestar.testing import TestClient

    with TestClient(app) as client:
        return client


def test_middleware_requests_ok(client: TestClient, mocker: MockerFixture):
    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")

    response = client.get("/foo/")
    assert response.status_code == 200
    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["consumer"] == "test1"
    assert mock.call_args.kwargs["method"] == "GET"
    assert mock.call_args.kwargs["path"] == "/foo"
    assert mock.call_args.kwargs["status_code"] == 200
    assert mock.call_args.kwargs["response_time"] > 0

    response = client.get("/foo/123")
    assert response.status_code == 200
    assert mock.call_count == 2
    assert mock.call_args is not None
    assert mock.call_args.kwargs["consumer"] == "test2"
    assert mock.call_args.kwargs["path"] == "/foo/{bar:str}"

    response = client.post("/bar")
    assert response.status_code == 201
    assert mock.call_count == 3
    assert mock.call_args is not None
    assert mock.call_args.kwargs["method"] == "POST"

    response = client.get("/stream")
    assert response.status_code == 200
    assert mock.call_count == 4
    assert mock.call_args is not None
    assert mock.call_args.kwargs["response_size"] == 6

    response = client.get("/schema/openapi.json")
    assert response.status_code == 200
    assert mock.call_count == 4  # OpenAPI paths are filtered out


def test_middleware_requests_unhandled(client: TestClient, mocker: MockerFixture):
    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")

    response = client.post("/xxx")
    assert response.status_code == 404
    mock.assert_not_called()


def test_middleware_requests_error(client: TestClient, mocker: MockerFixture):
    mock1 = mocker.patch("apitally.client.requests.RequestCounter.add_request")
    mock2 = mocker.patch("apitally.client.server_errors.ServerErrorCounter.add_server_error")

    response = client.post("/baz")
    assert response.status_code == 500
    mock1.assert_called_once()
    assert mock1.call_args is not None
    assert mock1.call_args.kwargs["method"] == "POST"
    assert mock1.call_args.kwargs["path"] == "/baz"
    assert mock1.call_args.kwargs["status_code"] == 500
    assert mock1.call_args.kwargs["response_time"] > 0

    mock2.assert_called_once()
    assert mock2.call_args is not None
    exception = mock2.call_args.kwargs["exception"]
    assert isinstance(exception, ValueError)


def test_middleware_validation_error(client: TestClient, mocker: MockerFixture):
    mock = mocker.patch("apitally.client.validation_errors.ValidationErrorCounter.add_validation_errors")

    # Validation error as foo must be an integer
    response = client.get("/val?foo=bar")
    assert response.status_code == 400

    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["method"] == "GET"
    assert mock.call_args.kwargs["path"] == "/val"
    assert len(mock.call_args.kwargs["detail"]) == 1
    assert mock.call_args.kwargs["detail"][0]["loc"] == ["query", "foo"]


def test_middleware_request_logging(client: TestClient, mocker: MockerFixture):
    from apitally.client.request_logging import BODY_TOO_LARGE

    mock = mocker.patch("apitally.client.request_logging.RequestLogger.log_request")

    response = client.get("/foo/123?foo=bar", headers={"Test-Header": "test"})
    assert response.status_code == 200
    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["method"] == "GET"
    assert mock.call_args.kwargs["request"]["path"] == "/foo/{bar:str}"
    assert mock.call_args.kwargs["request"]["url"] == "http://testserver.local/foo/123?foo=bar"
    assert ("test-header", "test") in mock.call_args.kwargs["request"]["headers"]
    assert mock.call_args.kwargs["request"]["consumer"] == "test2"
    assert mock.call_args.kwargs["response"]["status_code"] == 200
    assert mock.call_args.kwargs["response"]["response_time"] > 0
    assert ("content-type", "text/plain; charset=utf-8") in mock.call_args.kwargs["response"]["headers"]
    assert mock.call_args.kwargs["response"]["size"] > 0

    response = client.post("/bar", content=b"foo")
    assert response.status_code == 201
    assert mock.call_count == 2
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["method"] == "POST"
    assert mock.call_args.kwargs["request"]["path"] == "/bar"
    assert mock.call_args.kwargs["request"]["url"] == "http://testserver.local/bar"
    assert mock.call_args.kwargs["request"]["body"] == b"foo"
    assert mock.call_args.kwargs["response"]["body"] == b"bar: foo"

    mocker.patch("apitally.litestar.MAX_BODY_SIZE", 2)
    response = client.post("/bar", content=b"foo")
    assert response.status_code == 201
    assert mock.call_count == 3
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["body"] == BODY_TOO_LARGE
    assert mock.call_args.kwargs["response"]["body"] == BODY_TOO_LARGE


def test_get_startup_data(app: Litestar, mocker: MockerFixture):
    mock = mocker.patch("apitally.client.client_asyncio.ApitallyClient.set_startup_data")
    app.on_startup[0](app)  # type: ignore[call-arg]
    mock.assert_called_once()
    data = mock.call_args.args[0]
    assert len(data["openapi"]) > 0
    assert len(data["paths"]) == 6
    assert data["versions"]["litestar"]
    assert data["versions"]["app"] == "1.2.3"
    assert data["client"] == "python:litestar"



def test_request_counter():
    from apitally.client.requests import RequestCounter

    requests = RequestCounter()
    requests.add_request(
        consumer=None,
        method="GET",
        path="/test",
        status_code=200,
        response_time=0.105,
        request_size=None,
        response_size="123",
    )
    requests.add_request(
        consumer=None,
        method="GET",
        path="/test",
        status_code=200,
        response_time=0.227,
        request_size=None,
        response_size="321",
    )
    requests.add_request(
        consumer=None,
        method="POST",
        path="/test",
        status_code=204,
        response_time=0.1,
        request_size="123",
        response_size=None,
    )
    assert len(requests.request_counts) == 2

    data = requests.get_and_reset_requests()
    assert len(requests.request_counts) == 0
    assert len(data) == 2
    assert data[0]["method"] == "GET"
    assert data[0]["path"] == "/test"
    assert data[0]["status_code"] == 200
    assert data[0]["request_count"] == 2
    assert data[0]["request_size_sum"] == 0
    assert data[0]["response_size_sum"] > 0
    assert data[0]["response_times"][100] == 1
    assert data[0]["response_times"][220] == 1
    assert len(data[0]["request_sizes"]) == 0
    assert data[0]["response_sizes"][0] == 2
    assert data[1]["method"] == "POST"
    assert data[1]["request_size_sum"] > 0
    assert data[1]["response_size_sum"] == 0
    assert data[1]["request_sizes"][0] == 1



from __future__ import annotations

import base64
import gzip
import json
import time
from typing import TYPE_CHECKING, Iterator, Optional
from urllib.parse import quote

import pytest


if TYPE_CHECKING:
    from apitally.client.request_logging import RequestDict, RequestLogger, ResponseDict


@pytest.fixture()
def request_logger() -> Iterator[RequestLogger]:
    from apitally.client.request_logging import RequestLogger, RequestLoggingConfig

    config = RequestLoggingConfig(
        enabled=True,
        log_query_params=True,
        log_request_headers=True,
        log_request_body=True,
        log_response_headers=True,
        log_response_body=True,
    )
    request_logger = RequestLogger(config=config)
    yield request_logger
    request_logger.close()


@pytest.fixture()
def request_dict() -> RequestDict:
    return {
        "timestamp": time.time(),
        "method": "GET",
        "path": "/test",
        "url": "http://localhost:8000/test?foo=bar",
        "headers": [("Accept", "text/plain"), ("Content-Type", "text/plain")],
        "size": 100,
        "consumer": "test",
        "body": b"test",
    }


@pytest.fixture()
def response_dict() -> ResponseDict:
    return {
        "status_code": 200,
        "response_time": 0.1,
        "headers": [("Content-Type", "text/plain")],
        "size": 100,
        "body": b"test",
    }


async def test_request_logger_end_to_end(
    request_logger: RequestLogger, request_dict: RequestDict, response_dict: ResponseDict
):
    for _ in range(3):
        request_logger.log_request(request_dict, response_dict, RuntimeError("test"))

    request_logger.write_to_file()
    assert request_logger.current_file_size > 0

    request_logger.rotate_file()
    file = request_logger.get_file()
    assert file is not None

    compressed_data1 = b""
    async for chunk in file.stream_lines_compressed():
        compressed_data1 += chunk
    assert len(compressed_data1) > 0

    with file.open_compressed() as fp:
        compressed_data2 = fp.read()
    assert compressed_data1 == compressed_data2

    file.delete()

    decompressed_data = gzip.decompress(compressed_data1)
    json_lines = decompressed_data.decode("utf-8").strip().split("\n")
    assert len(json_lines) == 3

    for json_line in json_lines:
        item = json.loads(json_line)
        assert item["request"]["method"] == request_dict["method"]
        assert item["request"]["path"] == request_dict["path"]
        assert item["request"]["url"] == request_dict["url"]
        assert item["request"]["headers"] == [list(h) for h in request_dict["headers"]]
        assert item["request"]["size"] == request_dict["size"]
        assert item["request"]["consumer"] == request_dict["consumer"]
        assert item["response"]["status_code"] == response_dict["status_code"]
        assert item["response"]["response_time"] == response_dict["response_time"]
        assert item["response"]["headers"] == [list(h) for h in response_dict["headers"]]
        assert item["response"]["size"] == response_dict["size"]
        assert base64.b64decode(item["request"]["body"]) == request_dict["body"]
        assert base64.b64decode(item["response"]["body"]) == response_dict["body"]
        assert item["exception"]["type"] == "builtins.RuntimeError"
        assert item["exception"]["message"] == "test"


def test_request_log_exclusion(request_logger: RequestLogger, request_dict: RequestDict, response_dict: ResponseDict):
    request_logger.config.log_query_params = False
    request_logger.config.log_request_headers = False
    request_logger.config.log_request_body = False
    request_logger.config.log_response_headers = False
    request_logger.config.log_response_body = False
    request_logger.config.exclude_paths = ["/excluded$"]
    request_logger.config.exclude_callback = lambda _, response: response["status_code"] == 404

    request_logger.log_request(request_dict, response_dict)
    assert len(request_logger.write_deque) == 1
    item = request_logger.write_deque[0]
    assert item["request"]["url"] == "http://localhost:8000/test"
    assert "headers" not in item["request"]
    assert "body" not in item["request"]
    assert "headers" not in item["response"]
    assert "body" not in item["response"]

    response_dict["status_code"] = 404
    request_logger.log_request(request_dict, response_dict)
    assert len(request_logger.write_deque) == 1

    response_dict["status_code"] = 200
    request_dict["path"] = "/api/excluded"
    request_logger.log_request(request_dict, response_dict)
    assert len(request_logger.write_deque) == 1

    request_dict["path"] = "/healthz"
    request_logger.log_request(request_dict, response_dict)
    assert len(request_logger.write_deque) == 1

    request_dict["path"] = "/"
    request_dict["headers"] = [("User-Agent", "ELB-HealthChecker/2.0")]
    request_logger.log_request(request_dict, response_dict)
    assert len(request_logger.write_deque) == 1


def test_request_log_masking(request_logger: RequestLogger, request_dict: RequestDict, response_dict: ResponseDict):
    from apitally.client.request_logging import BODY_MASKED, MASKED

    MASKED_QUOTED = quote(MASKED)

    def mask_request_body_callback(request: RequestDict) -> Optional[bytes]:
        if request["method"] == "GET" and request["path"] == "/test":
            return None
        return request["body"]

    def mask_response_body_callback(request: RequestDict, response: ResponseDict) -> Optional[bytes]:
        if request["method"] == "GET" and request["path"] == "/test":
            return None
        return response["body"]

    request_logger.config.mask_headers = ["test"]
    request_logger.config.mask_query_params = ["test"]
    request_logger.config.mask_request_body_callback = mask_request_body_callback
    request_logger.config.mask_response_body_callback = mask_response_body_callback
    request_dict["url"] = "http://localhost/test?secret=123456&test=123456&other=abcdef"
    request_dict["headers"] += [("Authorization", "Bearer 123456"), ("X-Test", "123456")]
    request_logger.log_request(request_dict, response_dict)

    item = request_logger.write_deque[0]
    assert item["request"]["url"] == f"http://localhost/test?secret={MASKED_QUOTED}&test={MASKED_QUOTED}&other=abcdef"
    assert ("Authorization", "Bearer 123456") not in item["request"]["headers"]
    assert ("Authorization", MASKED) in item["request"]["headers"]
    assert ("X-Test", "123456") not in item["request"]["headers"]
    assert ("X-Test", MASKED) in item["request"]["headers"]
    assert ("Accept", "text/plain") in item["request"]["headers"]
    assert item["request"]["body"] == BODY_MASKED
    assert item["response"]["body"] == BODY_MASKED



from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING

import pytest
from pytest_mock import MockerFixture

from .constants import CLIENT_ID, ENV


if find_spec("flask") is None:
    pytest.skip("flask is not available", allow_module_level=True)

if TYPE_CHECKING:
    from flask import Flask


@pytest.fixture(scope="module")
def app(module_mocker: MockerFixture) -> Flask:
    from flask import Flask, g, request

    from apitally.flask import ApitallyMiddleware, RequestLoggingConfig

    module_mocker.patch("apitally.client.client_threading.ApitallyClient._instance", None)
    module_mocker.patch("apitally.client.client_threading.ApitallyClient.start_sync_loop")
    module_mocker.patch("apitally.client.client_threading.ApitallyClient.set_startup_data")
    module_mocker.patch("apitally.flask.ApitallyMiddleware.delayed_set_startup_data")

    app = Flask("test")
    app.wsgi_app = ApitallyMiddleware(  # type: ignore[method-assign]
        app,
        client_id=CLIENT_ID,
        env=ENV,
        request_logging_config=RequestLoggingConfig(
            enabled=True,
            log_request_body=True,
            log_response_body=True,
        ),
    )

    @app.route("/foo/<bar>")
    def foo_bar(bar: int):
        g.apitally_consumer = "test"
        return f"foo: {bar}", {"Content-Type": "text/plain"}

    @app.route("/bar", methods=["POST"])
    def bar():
        body = request.get_data()
        return "bar: " + body.decode(), {"Content-Type": "text/plain"}

    @app.route("/baz", methods=["PUT"])
    def baz():
        raise ValueError("baz")

    return app


def test_middleware_requests_ok(app: Flask, mocker: MockerFixture):
    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")
    client = app.test_client()

    response = client.get("/foo/123")
    assert response.status_code == 200
    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["consumer"] == "test"
    assert mock.call_args.kwargs["method"] == "GET"
    assert mock.call_args.kwargs["path"] == "/foo/<bar>"
    assert mock.call_args.kwargs["status_code"] == 200
    assert mock.call_args.kwargs["response_time"] > 0
    assert mock.call_args.kwargs["response_size"] > 0

    response = client.post("/bar")
    assert response.status_code == 200
    assert mock.call_count == 2
    assert mock.call_args is not None
    assert mock.call_args.kwargs["method"] == "POST"


def test_middleware_requests_error(app: Flask, mocker: MockerFixture):
    mock1 = mocker.patch("apitally.client.requests.RequestCounter.add_request")
    mock2 = mocker.patch("apitally.client.server_errors.ServerErrorCounter.add_server_error")
    client = app.test_client()

    response = client.put("/baz")
    assert response.status_code == 500
    mock1.assert_called_once()
    assert mock1.call_args is not None
    assert mock1.call_args.kwargs["method"] == "PUT"
    assert mock1.call_args.kwargs["path"] == "/baz"
    assert mock1.call_args.kwargs["status_code"] == 500
    assert mock1.call_args.kwargs["response_time"] > 0

    mock2.assert_called_once()
    assert mock2.call_args is not None
    exception = mock2.call_args.kwargs["exception"]
    assert isinstance(exception, ValueError)


def test_middleware_requests_unhandled(app: Flask, mocker: MockerFixture):
    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")
    client = app.test_client()

    response = client.post("/xxx")
    assert response.status_code == 404
    mock.assert_not_called()


def test_middleware_request_logging(app: Flask, mocker: MockerFixture):
    from apitally.client.request_logging import BODY_TOO_LARGE

    mock = mocker.patch("apitally.client.request_logging.RequestLogger.log_request")
    client = app.test_client()

    response = client.get("/foo/123?foo=bar", headers={"Test-Header": "test"})
    assert response.status_code == 200
    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["method"] == "GET"
    assert mock.call_args.kwargs["request"]["path"] == "/foo/<bar>"
    assert mock.call_args.kwargs["request"]["url"] == "http://localhost/foo/123?foo=bar"
    assert ("Test-Header", "test") in mock.call_args.kwargs["request"]["headers"]
    assert mock.call_args.kwargs["response"]["status_code"] == 200
    assert mock.call_args.kwargs["response"]["response_time"] > 0
    assert ("Content-Type", "text/plain") in mock.call_args.kwargs["response"]["headers"]
    assert mock.call_args.kwargs["response"]["size"] > 0

    response = client.post("/bar", data=b"foo")
    assert response.status_code == 200
    assert mock.call_count == 2
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["method"] == "POST"
    assert mock.call_args.kwargs["request"]["path"] == "/bar"
    assert mock.call_args.kwargs["request"]["url"] == "http://localhost/bar"
    assert mock.call_args.kwargs["request"]["body"] == b"foo"
    assert mock.call_args.kwargs["response"]["body"] == b"bar: foo"

    mocker.patch("apitally.flask.MAX_BODY_SIZE", 2)
    response = client.post("/bar", data=b"foo")
    assert response.status_code == 200
    assert mock.call_count == 3
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["body"] == BODY_TOO_LARGE
    assert mock.call_args.kwargs["response"]["body"] == BODY_TOO_LARGE


def test_get_startup_data(app: Flask):
    from apitally.flask import _get_startup_data

    data = _get_startup_data(app, app_version="1.2.3", openapi_url="/openapi.json")
    assert len(data["paths"]) == 3
    assert data["versions"]["flask"]
    assert data["versions"]["app"] == "1.2.3"
    assert data["client"] == "python:flask"



from __future__ import annotations

import time
from io import BytesIO
from threading import Timer
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Tuple
from warnings import warn

from flask import Flask, g
from flask.wrappers import Request, Response
from werkzeug.datastructures import Headers
from werkzeug.exceptions import NotFound
from werkzeug.test import Client

from apitally.client.client_threading import ApitallyClient
from apitally.client.consumers import Consumer as ApitallyConsumer
from apitally.client.request_logging import (
    BODY_TOO_LARGE,
    MAX_BODY_SIZE,
    RequestLogger,
    RequestLoggingConfig,
)
from apitally.common import get_versions


if TYPE_CHECKING:
    from _typeshed.wsgi import StartResponse, WSGIApplication, WSGIEnvironment
    from werkzeug.routing.map import Map


__all__ = ["ApitallyMiddleware", "ApitallyConsumer", "RequestLoggingConfig"]


class ApitallyMiddleware:
    def __init__(
        self,
        app: Flask,
        client_id: str,
        env: str = "dev",
        request_logging_config: Optional[RequestLoggingConfig] = None,
        app_version: Optional[str] = None,
        openapi_url: Optional[str] = None,
        proxy: Optional[str] = None,
    ) -> None:
        self.app = app
        self.wsgi_app = app.wsgi_app
        self.patch_handle_exception()
        self.client = ApitallyClient(
            client_id=client_id,
            env=env,
            request_logging_config=request_logging_config,
            proxy=proxy,
        )
        self.delayed_set_startup_data(app_version, openapi_url)

        self.capture_request_body = (
            self.client.request_logger.config.enabled and self.client.request_logger.config.log_request_body
        )
        self.capture_response_body = (
            self.client.request_logger.config.enabled and self.client.request_logger.config.log_response_body
        )

    def delayed_set_startup_data(self, app_version: Optional[str] = None, openapi_url: Optional[str] = None) -> None:
        # Short delay to allow app routes to be registered first
        timer = Timer(
            1.0,
            self._delayed_set_startup_data,
            kwargs={"app_version": app_version, "openapi_url": openapi_url},
        )
        timer.start()

    def _delayed_set_startup_data(self, app_version: Optional[str] = None, openapi_url: Optional[str] = None) -> None:
        data = _get_startup_data(self.app, app_version, openapi_url)
        self.client.set_startup_data(data)
        self.client.start_sync_loop()

    def __call__(self, environ: WSGIEnvironment, start_response: StartResponse) -> Iterable[bytes]:
        if not self.client.enabled:
            return self.wsgi_app(environ, start_response)

        timestamp = time.time()
        response_headers = Headers([])
        status_code = 0

        def catching_start_response(status: str, headers: List[Tuple[str, str]], exc_info=None):
            nonlocal status_code, response_headers
            status_code = int(status.split(" ")[0])
            response_headers = Headers(headers)
            return start_response(status, headers, exc_info)

        with self.app.app_context():
            request = Request(environ, populate_request=False, shallow=True)
            request_size = request.content_length
            request_body = b""
            if self.capture_request_body:
                request_body = (
                    _read_request_body(environ)
                    if request_size is not None and request_size <= MAX_BODY_SIZE
                    else BODY_TOO_LARGE
                )

            start_time = time.perf_counter()
            response = self.wsgi_app(environ, catching_start_response)
            response_time = time.perf_counter() - start_time

            response_body = b""
            response_content_type = response_headers.get("Content-Type")
            if self.capture_response_body and RequestLogger.is_supported_content_type(response_content_type):
                response_size = response_headers.get("Content-Length", type=int)
                if response_size is not None and response_size > MAX_BODY_SIZE:
                    response_body = BODY_TOO_LARGE
                else:
                    for chunk in response:
                        response_body += chunk
                        if len(response_body) > MAX_BODY_SIZE:
                            response_body = BODY_TOO_LARGE
                            break

            self.add_request(
                timestamp=timestamp,
                request=request,
                request_body=request_body,
                status_code=status_code,
                response_time=response_time,
                response_headers=response_headers,
                response_body=response_body,
            )
        return response

    def patch_handle_exception(self) -> None:
        original_handle_exception = self.app.handle_exception

        def handle_exception(e: Exception) -> Response:
            g.unhandled_exception = e
            return original_handle_exception(e)

        self.app.handle_exception = handle_exception  # type: ignore[method-assign]

    def add_request(
        self,
        timestamp: float,
        request: Request,
        request_body: bytes,
        status_code: int,
        response_time: float,
        response_headers: Headers,
        response_body: bytes,
    ) -> None:
        path = self.get_path(request.environ)
        response_size = response_headers.get("Content-Length", type=int)

        consumer = self.get_consumer()
        consumer_identifier = consumer.identifier if consumer else None
        self.client.consumer_registry.add_or_update_consumer(consumer)

        if path is not None and request.method != "OPTIONS":
            self.client.request_counter.add_request(
                consumer=consumer_identifier,
                method=request.method,
                path=path,
                status_code=status_code,
                response_time=response_time,
                request_size=request.content_length,
                response_size=response_size,
            )
            if status_code == 500 and "unhandled_exception" in g:
                self.client.server_error_counter.add_server_error(
                    consumer=consumer_identifier,
                    method=request.method,
                    path=path,
                    exception=g.unhandled_exception,
                )

        if self.client.request_logger.enabled:
            self.client.request_logger.log_request(
                request={
                    "timestamp": timestamp,
                    "method": request.method,
                    "path": path,
                    "url": request.url,
                    "headers": list(request.headers.items()),
                    "size": request.content_length,
                    "consumer": consumer_identifier,
                    "body": request_body,
                },
                response={
                    "status_code": status_code,
                    "response_time": response_time,
                    "headers": list(response_headers.items()),
                    "size": response_size,
                    "body": response_body,
                },
                exception=g.unhandled_exception if "unhandled_exception" in g else None,
            )

    def get_path(self, environ: WSGIEnvironment) -> Optional[str]:
        url_adapter = self.app.url_map.bind_to_environ(environ)
        try:
            endpoint, _ = url_adapter.match()
            rule = self.app.url_map._rules_by_endpoint[endpoint][0]
            return rule.rule
        except NotFound:
            return None

    def get_consumer(self) -> Optional[ApitallyConsumer]:
        if "apitally_consumer" in g and g.apitally_consumer:
            return ApitallyConsumer.from_string_or_object(g.apitally_consumer)
        if "consumer_identifier" in g and g.consumer_identifier:
            # Keeping this for legacy support
            warn(
                "Providing a consumer identifier via `g.consumer_identifier` is deprecated, "
                "use `g.apitally_consumer` instead.",
                DeprecationWarning,
            )
            return ApitallyConsumer.from_string_or_object(g.consumer_identifier)
        return None


def _get_startup_data(
    app: Flask, app_version: Optional[str] = None, openapi_url: Optional[str] = None
) -> Dict[str, Any]:
    data: Dict[str, Any] = {}
    if openapi_url and (openapi := _get_openapi(app, openapi_url)):
        data["openapi"] = openapi
    if paths := _get_paths(app.url_map):
        data["paths"] = paths
    data["versions"] = get_versions("flask", app_version=app_version)
    data["client"] = "python:flask"
    return data


def _get_paths(url_map: Map) -> List[Dict[str, str]]:
    return [
        {"path": rule.rule, "method": method}
        for rule in url_map.iter_rules()
        if rule.methods is not None and rule.rule != "/static/<path:filename>"
        for method in rule.methods
        if method not in ["HEAD", "OPTIONS"]
    ]


def _get_openapi(app: WSGIApplication, openapi_url: str) -> Optional[str]:
    client = Client(app)
    response = client.get(openapi_url)
    if response.status_code != 200:
        return None
    return response.get_data(as_text=True)


def _read_request_body(environ: WSGIEnvironment) -> bytes:
    length = int(environ.get("CONTENT_LENGTH", "0"))
    body = environ["wsgi.input"].read(length)
    environ["wsgi.input"] = BytesIO(body)
    return body



from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING, Optional

import pytest
from pytest_mock import MockerFixture

from .constants import CLIENT_ID, ENV


if find_spec("fastapi") is None:
    pytest.skip("fastapi is not available", allow_module_level=True)

if TYPE_CHECKING:
    from fastapi import FastAPI

# Global imports to avoid NameErrors during FastAPI dependency injection
try:
    from fastapi import Request
except ImportError:
    pass


@pytest.fixture(scope="module")
def app(module_mocker: MockerFixture) -> FastAPI:
    from fastapi import FastAPI

    from apitally.fastapi import ApitallyMiddleware

    module_mocker.patch("apitally.client.client_asyncio.ApitallyClient._instance", None)
    module_mocker.patch("apitally.client.client_asyncio.ApitallyClient.start_sync_loop")

    def identify_consumer(request: Request) -> Optional[str]:
        if consumer := request.query_params.get("consumer"):
            return consumer
        return None

    app = FastAPI()
    app.add_middleware(ApitallyMiddleware, client_id=CLIENT_ID, env=ENV, identify_consumer_callback=identify_consumer)

    @app.get("/foo/")
    def foo():
        return "foo"

    @app.get("/bar/")
    def bar():
        return "bar"

    @app.get("/baz/")
    def baz(request: Request):
        request.state.apitally_consumer = "baz"
        return "baz"

    return app


def test_get_openapi(app: FastAPI):
    from apitally.starlette import _get_openapi

    openapi = _get_openapi(app, "/openapi.json")
    assert openapi is not None
    assert len(openapi) > 0



import base64
import gzip
import re
import tempfile
import threading
import time
from collections import deque
from dataclasses import dataclass, field
from functools import lru_cache
from io import BufferedReader
from pathlib import Path
from typing import Any, AsyncIterator, Callable, Dict, List, Mapping, Optional, Tuple, TypedDict
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse
from uuid import uuid4

from apitally.client.logging import get_logger
from apitally.client.sentry import get_sentry_event_id_async
from apitally.client.server_errors import (
    get_exception_type,
    get_truncated_exception_msg,
    get_truncated_exception_traceback,
)


logger = get_logger(__name__)

MAX_BODY_SIZE = 50_000  # 50 KB (uncompressed)
MAX_FILE_SIZE = 1_000_000  # 1 MB (compressed)
MAX_REQUESTS_IN_DEQUE = 100  # Written to file every second, so limits logging to 100 rps
MAX_FILES_IN_DEQUE = 50
BODY_TOO_LARGE = b"<body too large>"
BODY_MASKED = b"<masked>"
MASKED = "******"
ALLOWED_CONTENT_TYPES = ["application/json", "text/plain"]
EXCLUDE_PATH_PATTERNS = [
    r"/_?healthz?$",
    r"/_?health[\-_]?checks?$",
    r"/_?heart[\-_]?beats?$",
    r"/ping$",
    r"/ready$",
    r"/live$",
]
EXCLUDE_USER_AGENT_PATTERNS = [
    r"health[\-_ ]?check",
    r"microsoft-azure-application-lb",
    r"googlehc",
    r"kube-probe",
]
MASK_QUERY_PARAM_PATTERNS = [
    r"auth",
    r"api-?key",
    r"secret",
    r"token",
    r"password",
    r"pwd",
]
MASK_HEADER_PATTERNS = [
    r"auth",
    r"api-?key",
    r"secret",
    r"token",
    r"cookie",
]


class RequestDict(TypedDict):
    timestamp: float
    method: str
    path: Optional[str]
    url: str
    headers: List[Tuple[str, str]]
    size: Optional[int]
    consumer: Optional[str]
    body: Optional[bytes]


class ResponseDict(TypedDict):
    status_code: int
    response_time: float
    headers: List[Tuple[str, str]]
    size: Optional[int]
    body: Optional[bytes]


@dataclass
class RequestLoggingConfig:
    """
    Configuration for request logging.

    Attributes:
        enabled: Whether request logging is enabled
        log_query_params: Whether to log query parameter values
        log_request_headers: Whether to log request header values
        log_request_body: Whether to log the request body (only if JSON or plain text)
        log_response_headers: Whether to log response header values
        log_response_body: Whether to log the response body (only if JSON or plain text)
        log_exception: Whether to log unhandled exceptions in case of server errors
        mask_query_params: Query parameter names to mask in logs. Expects regular expressions.
        mask_headers: Header names to mask in logs. Expects regular expressions.
        mask_request_body_callback: Callback to mask the request body. Expects (method, path, body) and returns the masked body as bytes or None.
        mask_response_body_callback: Callback to mask the response body. Expects (method, path, body) and returns the masked body as bytes or None.
        exclude_paths: Paths to exclude from logging. Expects regular expressions.
        exclude_callback: Callback to exclude requests from logging. Should expect two arguments, `request: RequestDict` and `response: ResponseDict`, and return True to exclude the request.
    """

    enabled: bool = False
    log_query_params: bool = True
    log_request_headers: bool = False
    log_request_body: bool = False
    log_response_headers: bool = True
    log_response_body: bool = False
    log_exception: bool = True
    mask_query_params: List[str] = field(default_factory=list)
    mask_headers: List[str] = field(default_factory=list)
    mask_request_body_callback: Optional[Callable[[RequestDict], Optional[bytes]]] = None
    mask_response_body_callback: Optional[Callable[[RequestDict, ResponseDict], Optional[bytes]]] = None
    exclude_paths: List[str] = field(default_factory=list)
    exclude_callback: Optional[Callable[[RequestDict, ResponseDict], bool]] = None


class TempGzipFile:
    def __init__(self) -> None:
        self.uuid = uuid4()
        self.file = tempfile.NamedTemporaryFile(
            suffix=".gz",
            prefix="apitally-",
            delete=False,
        )
        self.gzip_file = gzip.open(self.file, "wb")

    @property
    def path(self) -> Path:
        return Path(self.file.name)

    @property
    def size(self) -> int:
        return self.file.tell()

    def write_line(self, data: bytes) -> None:
        self.gzip_file.write(data + b"\n")

    def open_compressed(self) -> BufferedReader:
        return open(self.path, "rb")

    async def stream_lines_compressed(self) -> AsyncIterator[bytes]:
        with open(self.path, "rb") as fp:
            for line in fp:
                yield line

    def close(self) -> None:
        self.gzip_file.close()
        self.file.close()

    def delete(self) -> None:
        self.close()
        self.path.unlink(missing_ok=True)


class RequestLogger:
    def __init__(self, config: Optional[RequestLoggingConfig]) -> None:
        self.config = config or RequestLoggingConfig()
        self.enabled = self.config.enabled and _check_writable_fs()
        self.serialize = _get_json_serializer()
        self.write_deque: deque[Dict[str, Any]] = deque([], MAX_REQUESTS_IN_DEQUE)
        self.file_deque: deque[TempGzipFile] = deque([])
        self.file: Optional[TempGzipFile] = None
        self.lock = threading.Lock()
        self.suspend_until: Optional[float] = None

    @property
    def current_file_size(self) -> int:
        return self.file.size if self.file is not None else 0

    def log_request(
        self, request: RequestDict, response: ResponseDict, exception: Optional[BaseException] = None
    ) -> None:
        if not self.enabled or self.suspend_until is not None:
            return
        parsed_url = urlparse(request["url"])
        user_agent = self._get_user_agent(request["headers"])
        if (
            self._should_exclude_path(request["path"] or parsed_url.path)
            or self._should_exclude_user_agent(user_agent)
            or self._should_exclude(request, response)
        ):
            return

        query = self._mask_query_params(parsed_url.query) if self.config.log_query_params else ""
        request["url"] = urlunparse(parsed_url._replace(query=query))

        if not self.config.log_request_body or not self._has_supported_content_type(request["headers"]):
            request["body"] = None
        elif (
            self.config.mask_request_body_callback is not None
            and request["body"] is not None
            and request["body"] != BODY_TOO_LARGE
        ):
            try:
                request["body"] = self.config.mask_request_body_callback(request)
            except Exception:  # pragma: no cover
                logger.exception("User-provided mask_request_body_callback function raised an exception")
                request["body"] = None
            if request["body"] is None:
                request["body"] = BODY_MASKED
        if request["body"] is not None and len(request["body"]) > MAX_BODY_SIZE:
            request["body"] = BODY_TOO_LARGE

        if not self.config.log_response_body or not self._has_supported_content_type(response["headers"]):
            response["body"] = None
        elif (
            self.config.mask_response_body_callback is not None
            and response["body"] is not None
            and response["body"] != BODY_TOO_LARGE
        ):
            try:
                response["body"] = self.config.mask_response_body_callback(request, response)
            except Exception:  # pragma: no cover
                logger.exception("User-provided mask_response_body_callback function raised an exception")
                response["body"] = None
            if response["body"] is None:
                response["body"] = BODY_MASKED
        if response["body"] is not None and len(response["body"]) > MAX_BODY_SIZE:
            response["body"] = BODY_TOO_LARGE

        request["headers"] = self._mask_headers(request["headers"]) if self.config.log_request_headers else []
        response["headers"] = self._mask_headers(response["headers"]) if self.config.log_response_headers else []

        if request["size"] is not None and request["size"] < 0:
            request["size"] = None
        if response["size"] is not None and response["size"] < 0:
            response["size"] = None

        item: Dict[str, Any] = {
            "uuid": str(uuid4()),
            "request": _skip_empty_values(request),
            "response": _skip_empty_values(response),
        }
        if exception is not None and self.config.log_exception:
            item["exception"] = {
                "type": get_exception_type(exception),
                "message": get_truncated_exception_msg(exception),
                "traceback": get_truncated_exception_traceback(exception),
            }
            get_sentry_event_id_async(lambda event_id: item["exception"].update({"sentry_event_id": event_id}))
        self.write_deque.append(item)

    def write_to_file(self) -> None:
        if not self.enabled or len(self.write_deque) == 0:
            return
        with self.lock:
            if self.file is None:
                self.file = TempGzipFile()
            while True:
                try:
                    item = self.write_deque.popleft()
                    self.file.write_line(self.serialize(item))
                except IndexError:
                    break

    def get_file(self) -> Optional[TempGzipFile]:
        try:
            return self.file_deque.popleft()
        except IndexError:
            return None

    def retry_file_later(self, file: TempGzipFile) -> None:
        self.file_deque.appendleft(file)

    def rotate_file(self) -> None:
        if self.file is not None:
            with self.lock:
                self.file.close()
                self.file_deque.append(self.file)
                self.file = None

    def maintain(self) -> None:
        if self.current_file_size > MAX_FILE_SIZE:
            self.rotate_file()
        while len(self.file_deque) > MAX_FILES_IN_DEQUE:
            file = self.file_deque.popleft()
            file.delete()
        if self.suspend_until is not None and self.suspend_until < time.time():
            self.suspend_until = None

    def clear(self) -> None:
        self.write_deque.clear()
        self.rotate_file()
        for file in self.file_deque:
            file.delete()
        self.file_deque.clear()

    def close(self) -> None:
        self.enabled = False
        self.clear()

    def _should_exclude(self, request: RequestDict, response: ResponseDict) -> bool:
        if self.config.exclude_callback is not None:
            return self.config.exclude_callback(request, response)
        return False

    @lru_cache(maxsize=1000)
    def _should_exclude_path(self, url_path: str) -> bool:
        patterns = self.config.exclude_paths + EXCLUDE_PATH_PATTERNS
        return self._match_patterns(url_path, patterns)

    @lru_cache(maxsize=1000)
    def _should_exclude_user_agent(self, user_agent: Optional[str]) -> bool:
        return self._match_patterns(user_agent, EXCLUDE_USER_AGENT_PATTERNS) if user_agent is not None else False

    def _mask_query_params(self, query: str) -> str:
        query_params = parse_qsl(query)
        masked_query_params = [(k, v if not self._should_mask_query_param(k) else MASKED) for k, v in query_params]
        return urlencode(masked_query_params)

    def _mask_headers(self, headers: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        return [(k, v if not self._should_mask_header(k) else MASKED) for k, v in headers]

    @lru_cache(maxsize=100)
    def _should_mask_query_param(self, query_param_name: str) -> bool:
        patterns = self.config.mask_query_params + MASK_QUERY_PARAM_PATTERNS
        return self._match_patterns(query_param_name, patterns)

    @lru_cache(maxsize=100)
    def _should_mask_header(self, header_name: str) -> bool:
        patterns = self.config.mask_headers + MASK_HEADER_PATTERNS
        return self._match_patterns(header_name, patterns)

    @staticmethod
    def _match_patterns(value: str, patterns: List[str]) -> bool:
        for pattern in patterns:
            if re.search(pattern, value, re.I) is not None:
                return True
        return False

    @staticmethod
    def _get_user_agent(headers: List[Tuple[str, str]]) -> Optional[str]:
        return next((v for k, v in headers if k.lower() == "user-agent"), None)

    @staticmethod
    def _has_supported_content_type(headers: List[Tuple[str, str]]) -> bool:
        content_type = next((v for k, v in headers if k.lower() == "content-type"), None)
        return RequestLogger.is_supported_content_type(content_type)

    @staticmethod
    def is_supported_content_type(content_type: Optional[str]) -> bool:
        return content_type is not None and any(content_type.startswith(t) for t in ALLOWED_CONTENT_TYPES)


def _check_writable_fs() -> bool:
    try:
        with tempfile.NamedTemporaryFile():
            return True
    except (IOError, OSError):  # pragma: no cover
        logger.error("Unable to create temporary file for request logging")
        return False


def _get_json_serializer() -> Callable[[Any], bytes]:
    def default(obj: Any) -> Any:
        if isinstance(obj, bytes):
            return base64.b64encode(obj).decode()
        raise TypeError  # pragma: no cover

    try:
        import orjson  # type: ignore

        def orjson_dumps(obj: Any) -> bytes:
            return orjson.dumps(obj, default=default)

        return orjson_dumps
    except ImportError:
        import json

        def json_dumps(obj: Any) -> bytes:
            return json.dumps(obj, separators=(",", ":"), default=default).encode()

        return json_dumps


def _skip_empty_values(data: Mapping) -> Dict:
    return {
        k: v for k, v in data.items() if v is not None and not (isinstance(v, (list, dict, bytes, str)) and len(v) == 0)
    }



import asyncio
import contextlib
from typing import Callable, Set


_tasks: Set[asyncio.Task] = set()


def get_sentry_event_id_async(cb: Callable[[str], None]) -> None:
    try:
        from sentry_sdk.hub import Hub
        from sentry_sdk.scope import Scope
    except ImportError:
        return  # pragma: no cover
    if not hasattr(Scope, "get_isolation_scope") or not hasattr(Scope, "_last_event_id"):
        # sentry-sdk < 2.2.0 is not supported
        return  # pragma: no cover
    if Hub.current.client is None:
        return  # sentry-sdk not initialized

    scope = Scope.get_isolation_scope()
    if event_id := scope._last_event_id:
        cb(event_id)
        return

    async def _wait_for_sentry_event_id(scope: Scope) -> None:
        i = 0
        while not (event_id := scope._last_event_id) and i < 100:
            i += 1
            await asyncio.sleep(0.001)
        if event_id:
            cb(event_id)

    with contextlib.suppress(RuntimeError):  # ignore no running loop
        loop = asyncio.get_running_loop()
        task = loop.create_task(_wait_for_sentry_event_id(scope))
        _tasks.add(task)
        task.add_done_callback(_tasks.discard)



from __future__ import annotations

import contextlib
import threading
from collections import Counter
from dataclasses import dataclass
from math import floor
from typing import Any, Dict, List, Optional


@dataclass(frozen=True)
class RequestInfo:
    consumer: Optional[str]
    method: str
    path: str
    status_code: int


class RequestCounter:
    def __init__(self) -> None:
        self.request_counts: Counter[RequestInfo] = Counter()
        self.request_size_sums: Counter[RequestInfo] = Counter()
        self.response_size_sums: Counter[RequestInfo] = Counter()
        self.response_times: Dict[RequestInfo, Counter[int]] = {}
        self.request_sizes: Dict[RequestInfo, Counter[int]] = {}
        self.response_sizes: Dict[RequestInfo, Counter[int]] = {}
        self._lock = threading.Lock()

    def add_request(
        self,
        consumer: Optional[str],
        method: str,
        path: str,
        status_code: int,
        response_time: float,
        request_size: str | int | None = None,
        response_size: str | int | None = None,
    ) -> None:
        request_info = RequestInfo(
            consumer=consumer,
            method=method.upper(),
            path=path,
            status_code=status_code,
        )
        response_time_ms_bin = int(floor(response_time / 0.01) * 10)  # In ms, rounded down to nearest 10ms
        with self._lock:
            self.request_counts[request_info] += 1
            self.response_times.setdefault(request_info, Counter())[response_time_ms_bin] += 1
            if request_size is not None:
                with contextlib.suppress(ValueError):
                    request_size = int(request_size)
                    if request_size >= 0:
                        request_size_kb_bin = request_size // 1000  # In KB, rounded down to nearest 1KB
                        self.request_size_sums[request_info] += request_size
                        self.request_sizes.setdefault(request_info, Counter())[request_size_kb_bin] += 1
            if response_size is not None:
                with contextlib.suppress(ValueError):
                    response_size = int(response_size)
                    if response_size >= 0:
                        response_size_kb_bin = response_size // 1000  # In KB, rounded down to nearest 1KB
                        self.response_size_sums[request_info] += response_size
                        self.response_sizes.setdefault(request_info, Counter())[response_size_kb_bin] += 1

    def get_and_reset_requests(self) -> List[Dict[str, Any]]:
        data: List[Dict[str, Any]] = []
        with self._lock:
            for request_info, count in self.request_counts.items():
                data.append(
                    {
                        "consumer": request_info.consumer,
                        "method": request_info.method,
                        "path": request_info.path,
                        "status_code": request_info.status_code,
                        "request_count": count,
                        "request_size_sum": self.request_size_sums.get(request_info, 0),
                        "response_size_sum": self.response_size_sums.get(request_info, 0),
                        "response_times": self.response_times.get(request_info) or Counter(),
                        "request_sizes": self.request_sizes.get(request_info) or Counter(),
                        "response_sizes": self.response_sizes.get(request_info) or Counter(),
                    }
                )
            self.request_counts.clear()
            self.request_size_sums.clear()
            self.response_size_sums.clear()
            self.response_times.clear()
            self.request_sizes.clear()
            self.response_sizes.clear()
        return data



from __future__ import annotations

import threading
from collections import Counter
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple


@dataclass(frozen=True)
class ValidationError:
    consumer: Optional[str]
    method: str
    path: str
    loc: Tuple[str, ...]
    msg: str
    type: str


class ValidationErrorCounter:
    def __init__(self) -> None:
        self.error_counts: Counter[ValidationError] = Counter()
        self._lock = threading.Lock()

    def add_validation_errors(
        self, consumer: Optional[str], method: str, path: str, detail: List[Dict[str, Any]]
    ) -> None:
        with self._lock:
            for error in detail:
                try:
                    validation_error = ValidationError(
                        consumer=consumer,
                        method=method.upper(),
                        path=path,
                        loc=tuple(str(loc) for loc in error["loc"]),
                        msg=error["msg"],
                        type=error["type"],
                    )
                    self.error_counts[validation_error] += 1
                except (KeyError, TypeError):  # pragma: no cover
                    pass

    def get_and_reset_validation_errors(self) -> List[Dict[str, Any]]:
        data: List[Dict[str, Any]] = []
        with self._lock:
            for validation_error, count in self.error_counts.items():
                data.append(
                    {
                        "consumer": validation_error.consumer,
                        "method": validation_error.method,
                        "path": validation_error.path,
                        "loc": validation_error.loc,
                        "msg": validation_error.msg,
                        "type": validation_error.type,
                        "error_count": count,
                    }
                )
            self.error_counts.clear()
        return data



from django.http import HttpRequest, HttpResponse
from django.urls import path
from django.views import View
from django.views.decorators.http import require_http_methods
from rest_framework.decorators import api_view
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView


class FooView(APIView):
    """Foo"""

    def get(self, request: Request) -> Response:
        return Response("foo")


class FooBarView(APIView):
    def get(self, request: Request, bar: int) -> Response:
        request._request.apitally_consumer = "test"  # type: ignore[attr-defined]
        return Response({"foo": bar})


@api_view(["POST"])
def bar(request: Request) -> Response:
    return Response({"bar": request.data["foo"]})


@api_view(["PUT"])
def baz(request: Request) -> Response:
    raise ValueError("baz")


class SimpleClassView(View):
    def get(self, request: HttpRequest, pk: int, *args, **kwargs) -> HttpResponse:
        return HttpResponse(f"Hello from a class-based Django view with pk={pk}")


@require_http_methods(["GET"])
def simple_view(request: HttpRequest, pk: int) -> HttpResponse:
    return HttpResponse("Hello from a regular Django view")


urlpatterns = [
    path("foo/", FooView.as_view()),
    path("foo/<int:bar>/", FooBarView.as_view()),
    path("bar/", bar),
    path("baz/", baz),
    path("class/<int:pk>/", SimpleClassView.as_view()),
    path("func/<int:pk>/", simple_view),
]



import time
from typing import Any, Awaitable, Callable, Dict, List, Optional, Tuple, Union

from blacksheep import Application, Headers, Request, Response
from blacksheep.server.openapi.v3 import Info, OpenAPIHandler, Operation
from blacksheep.server.routing import RouteMatch

from apitally.client.client_asyncio import ApitallyClient
from apitally.client.consumers import Consumer as ApitallyConsumer
from apitally.client.request_logging import (
    BODY_TOO_LARGE,
    MAX_BODY_SIZE,
    RequestLogger,
    RequestLoggingConfig,
)
from apitally.common import get_versions, parse_int


__all__ = ["use_apitally", "ApitallyMiddleware", "ApitallyConsumer", "RequestLoggingConfig"]


def use_apitally(
    app: Application,
    client_id: str,
    env: str = "dev",
    request_logging_config: Optional[RequestLoggingConfig] = None,
    app_version: Optional[str] = None,
    identify_consumer_callback: Optional[Callable[[Request], Union[str, ApitallyConsumer, None]]] = None,
) -> None:
    original_get_match = app.router.get_match

    def _wrapped_router_get_match(request: Request) -> Optional[RouteMatch]:
        match = original_get_match(request)
        if match is not None:
            setattr(request, "_route_pattern", match.pattern.decode())
        return match

    app.router.get_match = _wrapped_router_get_match  # type: ignore[assignment,method-assign]

    middleware = ApitallyMiddleware(
        app,
        client_id,
        env=env,
        request_logging_config=request_logging_config,
        app_version=app_version,
        identify_consumer_callback=identify_consumer_callback,
    )
    app.middlewares.append(middleware)


class ApitallyMiddleware:
    def __init__(
        self,
        app: Application,
        client_id: str,
        env: str = "dev",
        request_logging_config: Optional[RequestLoggingConfig] = None,
        app_version: Optional[str] = None,
        identify_consumer_callback: Optional[Callable[[Request], Union[str, ApitallyConsumer, None]]] = None,
    ) -> None:
        self.app = app
        self.app_version = app_version
        self.identify_consumer_callback = identify_consumer_callback
        self.client = ApitallyClient(
            client_id=client_id,
            env=env,
            request_logging_config=request_logging_config,
        )
        self.app.on_start += self.after_start
        self.app.on_stop += self.on_stop

        self.capture_request_body = (
            self.client.request_logger.config.enabled and self.client.request_logger.config.log_request_body
        )
        self.capture_response_body = (
            self.client.request_logger.config.enabled and self.client.request_logger.config.log_response_body
        )

    async def after_start(self, application: Application) -> None:
        data = _get_startup_data(application, app_version=self.app_version)
        self.client.set_startup_data(data)
        self.client.start_sync_loop()

    async def on_stop(self, application: Application) -> None:
        await self.client.handle_shutdown()

    def get_consumer(self, request: Request) -> Optional[ApitallyConsumer]:
        identity = request.user or request.identity or None
        if identity is not None and identity.has_claim("apitally_consumer"):
            return ApitallyConsumer.from_string_or_object(identity.get("apitally_consumer"))
        if self.identify_consumer_callback is not None:
            consumer = self.identify_consumer_callback(request)
            return ApitallyConsumer.from_string_or_object(consumer)
        return None

    async def __call__(self, request: Request, handler: Callable[[Request], Awaitable[Response]]) -> Response:
        if not self.client.enabled:
            return await handler(request)

        timestamp = time.time()
        start_time = time.perf_counter()
        response: Optional[Response] = None
        exception: Optional[BaseException] = None

        try:
            response = await handler(request)
        except BaseException as e:
            exception = e
            raise e from None
        finally:
            response_time = time.perf_counter() - start_time

            consumer = self.get_consumer(request)
            consumer_identifier = consumer.identifier if consumer else None
            self.client.consumer_registry.add_or_update_consumer(consumer)

            route_pattern: Optional[str] = getattr(request, "_route_pattern", None)
            request_size = parse_int(request.get_first_header(b"Content-Length"))
            request_content_type = (request.content_type() or b"").decode() or None
            request_body = b""

            response_status = response.status if response else 500
            response_size: Optional[int] = None
            response_headers = Headers()
            response_body = b""

            if self.capture_request_body and RequestLogger.is_supported_content_type(request_content_type):
                if request_size is not None and request_size > MAX_BODY_SIZE:
                    request_body = BODY_TOO_LARGE
                else:
                    request_body = await request.read() or b""
                    if request_size is None:
                        request_size = len(request_body)

            if response is not None:
                response_size = (
                    response.content.length
                    if response.content
                    else parse_int(response.get_first_header(b"Content-Length"))
                )
                response_content_type = (response.content_type() or b"").decode()

                response_headers = response.headers.clone()
                if not response_headers.contains(b"Content-Type") and response.content:
                    response_headers.set(b"Content-Type", response.content.type)
                if not response_headers.contains(b"Content-Length") and response.content:
                    response_headers.set(b"Content-Length", str(response.content.length).encode())

                if self.capture_response_body and RequestLogger.is_supported_content_type(response_content_type):
                    if response_size is not None and response_size > MAX_BODY_SIZE:
                        response_body = BODY_TOO_LARGE
                    else:
                        response_body = await response.read() or b""
                        if response_size is None or response_size < 0:
                            response_size = len(response_body)

            if route_pattern and request.method.upper() != "OPTIONS":
                self.client.request_counter.add_request(
                    consumer=consumer_identifier,
                    method=request.method.upper(),
                    path=route_pattern,
                    status_code=response_status,
                    response_time=response_time,
                    request_size=request_size,
                    response_size=response_size,
                )

                if response_status == 500 and exception is not None:
                    self.client.server_error_counter.add_server_error(
                        consumer=consumer_identifier,
                        method=request.method.upper(),
                        path=route_pattern,
                        exception=exception,
                    )

            if self.client.request_logger.enabled:
                self.client.request_logger.log_request(
                    request={
                        "timestamp": timestamp,
                        "method": request.method.upper(),
                        "path": route_pattern,
                        "url": _get_full_url(request),
                        "headers": _transform_headers(request.headers),
                        "size": request_size,
                        "consumer": consumer_identifier,
                        "body": request_body,
                    },
                    response={
                        "status_code": response_status,
                        "response_time": response_time,
                        "headers": _transform_headers(response_headers),
                        "size": response_size,
                        "body": response_body,
                    },
                    exception=exception,
                )

        return response


def _get_full_url(request: Request) -> str:
    return f"{request.scheme}://{request.host}/{str(request.url).lstrip('/')}"


def _transform_headers(headers: Headers) -> List[Tuple[str, str]]:
    return [(key.decode(), value.decode()) for key, value in headers.items()]


def _get_startup_data(app: Application, app_version: Optional[str] = None) -> Dict[str, Any]:
    return {
        "paths": _get_paths(app),
        "versions": get_versions("blacksheep", app_version=app_version),
        "client": "python:blacksheep",
    }


def _get_paths(app: Application) -> List[Dict[str, str]]:
    openapi = OpenAPIHandler(info=Info(title="", version=""))
    paths = []
    methods = ("get", "put", "post", "delete", "options", "head", "patch", "trace")
    for path, path_item in openapi.get_paths(app).items():
        for method in methods:
            operation: Operation = getattr(path_item, method, None)
            if operation is not None:
                item = {"method": method.upper(), "path": path}
                if operation.summary:
                    item["summary"] = operation.summary
                if operation.description:
                    item["description"] = operation.description
                paths.append(item)
    return paths



from __future__ import annotations

import json
import sys
from importlib.util import find_spec
from typing import TYPE_CHECKING

import pytest
from pytest_mock import MockerFixture


if find_spec("rest_framework") is None:
    pytest.skip("django-rest-framework is not available", allow_module_level=True)

if TYPE_CHECKING:
    from rest_framework.test import APIClient


@pytest.fixture(scope="module")
def reset_modules() -> None:
    for module in list(sys.modules):
        if module.startswith("django.") or module.startswith("rest_framework.") or module.startswith("apitally."):
            del sys.modules[module]


@pytest.fixture(scope="module", autouse=True)
def setup(reset_modules, module_mocker: MockerFixture) -> None:
    import django
    from django.conf import settings

    from apitally.django_rest_framework import RequestLoggingConfig

    module_mocker.patch("apitally.client.client_threading.ApitallyClient._instance", None)
    module_mocker.patch("apitally.client.client_threading.ApitallyClient.start_sync_loop")
    module_mocker.patch("apitally.client.client_threading.ApitallyClient.set_startup_data")
    module_mocker.patch("apitally.django.ApitallyMiddleware.config", None)

    settings.configure(
        ROOT_URLCONF="tests.django_rest_framework_urls",
        ALLOWED_HOSTS=["testserver"],
        SECRET_KEY="secret",
        MIDDLEWARE=[
            "apitally.django_rest_framework.ApitallyMiddleware",
            "django.middleware.common.CommonMiddleware",
        ],
        INSTALLED_APPS=[
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "rest_framework",
        ],
        APITALLY_MIDDLEWARE={
            "client_id": "76b5cb91-a0a4-4ea0-a894-57d2b9fcb2c9",
            "env": "dev",
            "request_logging_config": RequestLoggingConfig(
                enabled=True,
                log_request_body=True,
                log_response_body=True,
            ),
            "include_django_views": True,
            "urlconf": ["tests.django_rest_framework_urls"],
        },
    )
    django.setup()


@pytest.fixture(scope="module")
def client(module_mocker: MockerFixture) -> APIClient:
    import django
    from rest_framework.test import APIClient

    if django.VERSION[0] < 3:
        module_mocker.patch("django.test.client.Client.store_exc_info")  # Simulate raise_request_exception=False
    return APIClient(raise_request_exception=False)


def test_middleware_requests_ok(client: APIClient, mocker: MockerFixture):
    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")

    response = client.get("/foo/123/")
    assert response.status_code == 200
    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["consumer"] == "test"
    assert mock.call_args.kwargs["method"] == "GET"
    assert mock.call_args.kwargs["path"] == "/foo/{bar}/"
    assert mock.call_args.kwargs["status_code"] == 200
    assert mock.call_args.kwargs["response_time"] > 0
    assert int(mock.call_args.kwargs["response_size"]) > 0

    response = client.post("/bar/", data={"foo": "bar"})
    assert response.status_code == 200
    assert mock.call_count == 2
    assert mock.call_args is not None
    assert mock.call_args.kwargs["method"] == "POST"
    assert int(mock.call_args.kwargs["request_size"]) > 0

    response = client.get("/func/123/")
    assert response.status_code == 200
    assert mock.call_count == 3
    assert mock.call_args is not None
    assert mock.call_args.kwargs["method"] == "GET"
    assert mock.call_args.kwargs["path"] == "/func/{pk}/"


def test_middleware_requests_404(client: APIClient, mocker: MockerFixture):
    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")

    response = client.get("/api/none")
    assert response.status_code == 404
    mock.assert_not_called()


def test_middleware_requests_error(client: APIClient, mocker: MockerFixture):
    mock1 = mocker.patch("apitally.client.requests.RequestCounter.add_request")
    mock2 = mocker.patch("apitally.client.server_errors.ServerErrorCounter.add_server_error")

    response = client.put("/baz/")
    assert response.status_code == 500
    mock1.assert_called_once()
    assert mock1.call_args is not None
    assert mock1.call_args.kwargs["method"] == "PUT"
    assert mock1.call_args.kwargs["path"] == "/baz/"
    assert mock1.call_args.kwargs["status_code"] == 500
    assert mock1.call_args.kwargs["response_time"] > 0

    mock2.assert_called_once()
    assert mock2.call_args is not None
    exception = mock2.call_args.kwargs["exception"]
    assert isinstance(exception, ValueError)


def test_middleware_request_logging(client: APIClient, mocker: MockerFixture):
    from apitally.client.request_logging import BODY_TOO_LARGE

    mock = mocker.patch("apitally.client.request_logging.RequestLogger.log_request")

    response = client.get("/foo/123/?foo=bar", HTTP_TEST_HEADER="test")
    assert response.status_code == 200
    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["method"] == "GET"
    assert mock.call_args.kwargs["request"]["path"] == "/foo/{bar}/"
    assert mock.call_args.kwargs["request"]["url"] == "http://testserver/foo/123/?foo=bar"
    assert ("Test-Header", "test") in mock.call_args.kwargs["request"]["headers"]
    assert mock.call_args.kwargs["request"]["consumer"] == "test"
    assert mock.call_args.kwargs["response"]["status_code"] == 200
    assert mock.call_args.kwargs["response"]["response_time"] > 0
    assert ("Content-Type", "application/json") in mock.call_args.kwargs["response"]["headers"]
    assert mock.call_args.kwargs["response"]["size"] > 0

    response = client.post("/bar/", data={"foo": "foo"}, format="json")
    assert response.status_code == 200
    assert mock.call_count == 2
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["method"] == "POST"
    assert mock.call_args.kwargs["request"]["path"] == "/bar/"
    assert mock.call_args.kwargs["request"]["url"] == "http://testserver/bar/"
    assert mock.call_args.kwargs["request"]["body"] == b'{"foo":"foo"}'
    assert mock.call_args.kwargs["response"]["body"] == b'{"bar":"foo"}'

    mocker.patch("apitally.django.MAX_BODY_SIZE", 2)
    response = client.post("/bar/", data={"foo": "foo"}, format="json")
    assert response.status_code == 200
    assert mock.call_count == 3
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["body"] == BODY_TOO_LARGE
    assert mock.call_args.kwargs["response"]["body"] == BODY_TOO_LARGE


def test_get_startup_data():
    from apitally.django import _get_startup_data

    data = _get_startup_data(app_version="1.2.3", urlconfs=[None])
    openapi = json.loads(data["openapi"])
    assert len(data["paths"]) == 4
    assert len(openapi["paths"]) == 4

    assert data["versions"]["django"]
    assert data["versions"]["djangorestframework"]
    assert data["versions"]["app"] == "1.2.3"
    assert data["client"] == "python:django"


def test_get_paths():
    from apitally.django import _get_paths

    paths = _get_paths([None])
    assert len(paths) == 4
    assert paths[0]["method"] == "GET"
    assert paths[0]["path"] == "/foo/"
    assert paths[1]["path"] == "/foo/{bar}/"

    paths = _get_paths([None], include_django_views=True)
    assert len(paths) == 5  # Also includes Django class-based view, but not function-based view



import gzip
import json
import sys
from importlib.metadata import PackageNotFoundError, version
from typing import Any, Dict, Optional, Union


def parse_int(x: Union[str, bytes, int, None]) -> Optional[int]:
    if x is None:
        return None
    try:
        return int(x)
    except ValueError:
        return None


def try_json_loads(s: bytes, encoding: Optional[str] = None) -> Any:
    if encoding is not None and encoding.lower() == "gzip":
        try:
            s = gzip.decompress(s)
        except Exception:
            pass
    try:
        return json.loads(s)
    except Exception:
        return None


def get_versions(*packages, app_version: Optional[str] = None) -> Dict[str, str]:
    versions = _get_common_package_versions()
    for package in packages:
        versions[package] = _get_package_version(package)
    if app_version:
        versions["app"] = app_version
    return {n: v for n, v in versions.items() if v is not None}


def _get_common_package_versions() -> Dict[str, Optional[str]]:
    return {
        "python": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        "apitally": _get_package_version("apitally"),
        "uvicorn": _get_package_version("uvicorn"),
        "hypercorn": _get_package_version("hypercorn"),
        "daphne": _get_package_version("daphne"),
        "gunicorn": _get_package_version("gunicorn"),
        "uwsgi": _get_package_version("uwsgi"),
    }


def _get_package_version(name: str) -> Optional[str]:
    try:
        return version(name)
    except PackageNotFoundError:
        return None



from __future__ import annotations

import os
import re
import threading
import time
from abc import ABC
from typing import Any, Dict, Optional, Type, TypeVar, cast
from uuid import UUID, uuid4

from apitally.client.consumers import ConsumerRegistry
from apitally.client.logging import get_logger
from apitally.client.request_logging import RequestLogger, RequestLoggingConfig
from apitally.client.requests import RequestCounter
from apitally.client.server_errors import ServerErrorCounter
from apitally.client.validation_errors import ValidationErrorCounter


logger = get_logger(__name__)

HUB_BASE_URL = os.getenv("APITALLY_HUB_BASE_URL") or "https://hub.apitally.io"
HUB_VERSION = "v2"
REQUEST_TIMEOUT = 10
MAX_QUEUE_TIME = 3600
SYNC_INTERVAL = 60
INITIAL_SYNC_INTERVAL = 10
INITIAL_SYNC_INTERVAL_DURATION = 3600

TApitallyClient = TypeVar("TApitallyClient", bound="ApitallyClientBase")


class ApitallyClientBase(ABC):
    _instance: Optional[ApitallyClientBase] = None
    _lock = threading.Lock()

    def __new__(cls: Type[TApitallyClient], *args, **kwargs) -> TApitallyClient:
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
        return cast(TApitallyClient, cls._instance)

    def __init__(self, client_id: str, env: str, request_logging_config: Optional[RequestLoggingConfig] = None) -> None:
        if hasattr(self, "client_id"):
            raise RuntimeError("Apitally client is already initialized")  # pragma: no cover
        try:
            UUID(client_id)
        except ValueError:
            raise ValueError(f"invalid client_id '{client_id}' (expecting hexadecimal UUID format)")
        if re.match(r"^[\w-]{1,32}$", env) is None:
            raise ValueError(f"invalid env '{env}' (expecting 1-32 alphanumeric lowercase characters and hyphens only)")

        self.client_id = client_id
        self.env = env
        self.enabled = True
        self.instance_uuid = str(uuid4())
        self.request_counter = RequestCounter()
        self.validation_error_counter = ValidationErrorCounter()
        self.server_error_counter = ServerErrorCounter()
        self.consumer_registry = ConsumerRegistry()
        self.request_logger = RequestLogger(request_logging_config)

        self._startup_data: Optional[Dict[str, Any]] = None
        self._startup_data_sent = False
        self._started_at = time.time()

    @classmethod
    def get_instance(cls: Type[TApitallyClient]) -> TApitallyClient:
        if cls._instance is None:
            raise RuntimeError("Apitally client not initialized")  # pragma: no cover
        return cast(TApitallyClient, cls._instance)

    @property
    def sync_interval(self) -> float:
        return (
            SYNC_INTERVAL if time.time() - self._started_at > INITIAL_SYNC_INTERVAL_DURATION else INITIAL_SYNC_INTERVAL
        )

    @property
    def hub_url(self) -> str:
        return f"{HUB_BASE_URL}/{HUB_VERSION}/{self.client_id}/{self.env}"

    def add_uuids_to_data(self, data: Dict[str, Any]) -> Dict[str, Any]:
        data_with_uuids = {
            "instance_uuid": self.instance_uuid,
            "message_uuid": str(uuid4()),
        }
        data_with_uuids.update(data)
        return data_with_uuids

    def get_sync_data(self) -> Dict[str, Any]:
        data = {
            "timestamp": time.time(),
            "requests": self.request_counter.get_and_reset_requests(),
            "validation_errors": self.validation_error_counter.get_and_reset_validation_errors(),
            "server_errors": self.server_error_counter.get_and_reset_server_errors(),
            "consumers": self.consumer_registry.get_and_reset_updated_consumers(),
        }
        return self.add_uuids_to_data(data)



from __future__ import annotations

import contextlib
import json
import re
import time
from dataclasses import dataclass
from importlib import import_module
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Union
from warnings import warn

from django.conf import settings
from django.contrib.admindocs.views import extract_views_from_urlpatterns, simplify_regex
from django.urls import URLPattern, URLResolver, get_resolver
from django.utils.module_loading import import_string
from django.views.generic.base import View

from apitally.client.client_threading import ApitallyClient
from apitally.client.consumers import Consumer as ApitallyConsumer
from apitally.client.logging import get_logger
from apitally.client.request_logging import (
    BODY_TOO_LARGE,
    MAX_BODY_SIZE,
    RequestLogger,
    RequestLoggingConfig,
)
from apitally.common import get_versions, parse_int, try_json_loads


if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse
    from ninja import NinjaAPI


__all__ = ["ApitallyMiddleware", "ApitallyConsumer", "RequestLoggingConfig"]
logger = get_logger(__name__)


@dataclass
class ApitallyMiddlewareConfig:
    client_id: str
    env: str
    request_logging_config: Optional[RequestLoggingConfig]
    app_version: Optional[str]
    identify_consumer_callback: Optional[Callable[[HttpRequest], Union[str, ApitallyConsumer, None]]]
    include_django_views: bool
    urlconfs: List[Optional[str]]
    proxy: Optional[str]


class ApitallyMiddleware:
    config: Optional[ApitallyMiddlewareConfig] = None

    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]) -> None:
        self.get_response = get_response
        self.drf_available = _check_import("rest_framework")
        self.drf_endpoint_enumerator = None
        self.ninja_available = _check_import("ninja")
        self.include_django_views = False
        self.callbacks = set()

        if self.config is None:
            config = getattr(settings, "APITALLY_MIDDLEWARE", {})
            self.configure(**config)
            assert self.config is not None

        if self.drf_available:
            from rest_framework.schemas.generators import EndpointEnumerator

            self.drf_endpoint_enumerator = EndpointEnumerator()
            if None not in self.config.urlconfs:
                self.callbacks.update(_get_drf_callbacks(self.config.urlconfs))
        if self.ninja_available and None not in self.config.urlconfs:
            self.callbacks.update(_get_ninja_callbacks(self.config.urlconfs))
        if self.config.include_django_views:
            self.callbacks.update(_get_django_callbacks(self.config.urlconfs))
            self.include_django_views = True

        self.client = ApitallyClient(
            client_id=self.config.client_id,
            env=self.config.env,
            request_logging_config=self.config.request_logging_config,
            proxy=self.config.proxy,
        )
        self.client.set_startup_data(
            _get_startup_data(
                app_version=self.config.app_version,
                urlconfs=self.config.urlconfs,
            )
        )
        self.client.start_sync_loop()

        self.capture_request_body = (
            self.client.request_logger.config.enabled and self.client.request_logger.config.log_request_body
        )
        self.capture_response_body = (
            self.client.request_logger.config.enabled and self.client.request_logger.config.log_response_body
        )

    @classmethod
    def configure(
        cls,
        client_id: str,
        env: str = "dev",
        request_logging_config: Optional[RequestLoggingConfig] = None,
        app_version: Optional[str] = None,
        identify_consumer_callback: Optional[str] = None,
        include_django_views: bool = False,
        urlconf: Optional[Union[List[Optional[str]], str]] = None,
        proxy: Optional[str] = None,
    ) -> None:
        cls.config = ApitallyMiddlewareConfig(
            client_id=client_id,
            env=env,
            request_logging_config=request_logging_config,
            app_version=app_version,
            identify_consumer_callback=import_string(identify_consumer_callback)
            if identify_consumer_callback
            else None,
            include_django_views=include_django_views,
            urlconfs=[urlconf] if urlconf is None or isinstance(urlconf, str) else urlconf,
            proxy=proxy,
        )

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if self.client.enabled and request.method is not None and request.method != "OPTIONS":
            timestamp = time.time()

            request_size = parse_int(request.headers.get("Content-Length"))
            request_body = b""
            if self.capture_request_body:
                request_body = (
                    request.body
                    if request_size is not None and request_size <= MAX_BODY_SIZE and len(request.body) <= MAX_BODY_SIZE
                    else BODY_TOO_LARGE
                )

            start_time = time.perf_counter()
            response = self.get_response(request)
            response_time = time.perf_counter() - start_time
            path = self.get_path(request)

            if path is not None:
                try:
                    consumer = self.get_consumer(request)
                    consumer_identifier = consumer.identifier if consumer else None
                    self.client.consumer_registry.add_or_update_consumer(consumer)
                except Exception:  # pragma: no cover
                    logger.exception("Failed to get consumer for request")
                    consumer_identifier = None

                response_size = (
                    parse_int(response["Content-Length"])
                    if response.has_header("Content-Length")
                    else (len(response.content) if not response.streaming else None)
                )
                response_body = b""
                response_content_type = response.get("Content-Type")
                if (
                    self.capture_response_body
                    and not response.streaming
                    and RequestLogger.is_supported_content_type(response_content_type)
                ):
                    response_body = (
                        response.content
                        if response_size is not None and response_size <= MAX_BODY_SIZE
                        else BODY_TOO_LARGE
                    )

                try:
                    self.client.request_counter.add_request(
                        consumer=consumer_identifier,
                        method=request.method,
                        path=path,
                        status_code=response.status_code,
                        response_time=response_time,
                        request_size=request_size,
                        response_size=response_size,
                    )
                except Exception:  # pragma: no cover
                    logger.exception("Failed to log request metadata")

                if (
                    response.status_code == 422
                    and (content_type := response.get("Content-Type")) is not None
                    and content_type.startswith("application/json")
                ):
                    try:
                        body = try_json_loads(response.content, encoding=response.get("Content-Encoding"))
                        if isinstance(body, dict) and "detail" in body and isinstance(body["detail"], list):
                            # Log Django Ninja / Pydantic validation errors
                            self.client.validation_error_counter.add_validation_errors(
                                consumer=consumer_identifier,
                                method=request.method,
                                path=path,
                                detail=body["detail"],
                            )
                    except Exception:  # pragma: no cover
                        logger.exception("Failed to log validation errors")

                if response.status_code == 500 and hasattr(request, "unhandled_exception"):
                    try:
                        self.client.server_error_counter.add_server_error(
                            consumer=consumer_identifier,
                            method=request.method,
                            path=path,
                            exception=getattr(request, "unhandled_exception"),
                        )
                    except Exception:  # pragma: no cover
                        logger.exception("Failed to log server error")

                if self.client.request_logger.enabled:
                    self.client.request_logger.log_request(
                        request={
                            "timestamp": timestamp,
                            "method": request.method,
                            "path": path,
                            "url": request.build_absolute_uri(),
                            "headers": list(request.headers.items()),
                            "size": request_size,
                            "consumer": consumer_identifier,
                            "body": request_body,
                        },
                        response={
                            "status_code": response.status_code,
                            "response_time": response_time,
                            "headers": list(response.items()),
                            "size": response_size,
                            "body": response_body,
                        },
                        exception=getattr(request, "unhandled_exception", None),
                    )
        else:
            response = self.get_response(request)

        return response

    def process_exception(self, request: HttpRequest, exception: Exception) -> None:
        setattr(request, "unhandled_exception", exception)
        return None

    def get_path(self, request: HttpRequest) -> Optional[str]:
        if (match := request.resolver_match) is not None:
            try:
                if self.callbacks and match.func not in self.callbacks:
                    return None
                if self.drf_endpoint_enumerator is not None:
                    from rest_framework.schemas.generators import is_api_view

                    if is_api_view(match.func):
                        return self.drf_endpoint_enumerator.get_path_from_regex(match.route)
                if self.ninja_available:
                    from ninja.operation import PathView

                    if hasattr(match.func, "__self__") and isinstance(match.func.__self__, PathView):
                        return _transform_path(match.route)
                if self.include_django_views:
                    return _transform_path(match.route)
            except Exception:  # pragma: no cover
                logger.exception("Failed to get path for request")
        return None

    def get_consumer(self, request: HttpRequest) -> Optional[ApitallyConsumer]:
        if hasattr(request, "apitally_consumer") and request.apitally_consumer:
            return ApitallyConsumer.from_string_or_object(request.apitally_consumer)
        if hasattr(request, "consumer_identifier") and request.consumer_identifier:
            # Keeping this for legacy support
            warn(
                "Providing a consumer identifier via `request.consumer_identifier` is deprecated, "
                "use `request.apitally_consumer` instead.",
                DeprecationWarning,
            )
            return ApitallyConsumer.from_string_or_object(request.consumer_identifier)
        if self.config is not None and self.config.identify_consumer_callback is not None:
            consumer = self.config.identify_consumer_callback(request)
            return ApitallyConsumer.from_string_or_object(consumer)
        return None


def _get_startup_data(
    app_version: Optional[str], urlconfs: List[Optional[str]], include_django_views: bool = False
) -> Dict[str, Any]:
    data: Dict[str, Any] = {}
    try:
        data["paths"] = _get_paths(urlconfs, include_django_views=include_django_views)
    except Exception:  # pragma: no cover
        data["paths"] = []
        logger.exception("Failed to get paths")
    try:
        data["openapi"] = _get_openapi(urlconfs)
    except Exception:  # pragma: no cover
        logger.exception("Failed to get OpenAPI schema")
    data["versions"] = get_versions("django", "djangorestframework", "django-ninja", app_version=app_version)
    data["client"] = "python:django"
    return data


def _get_openapi(urlconfs: List[Optional[str]]) -> Optional[str]:
    drf_schema = None
    ninja_schema = None
    with contextlib.suppress(ImportError):
        drf_schema = _get_drf_schema(urlconfs)
    with contextlib.suppress(ImportError):
        ninja_schema = _get_ninja_schema(urlconfs)
    if drf_schema is not None and ninja_schema is None:
        return json.dumps(drf_schema)
    elif ninja_schema is not None and drf_schema is None:
        return json.dumps(ninja_schema)
    return None  # pragma: no cover


def _get_paths(urlconfs: List[Optional[str]], include_django_views: bool = False) -> List[Dict[str, str]]:
    paths = []
    with contextlib.suppress(ImportError):
        paths.extend(_get_drf_paths(urlconfs))
    with contextlib.suppress(ImportError):
        paths.extend(_get_ninja_paths(urlconfs))
    if include_django_views:
        paths.extend(_get_django_paths(urlconfs))
    return _deduplicate_paths(paths)


def _deduplicate_paths(paths: List[Dict[str, str]]) -> List[Dict[str, str]]:
    seen = set()
    deduplicated_paths = []
    for path in paths:
        key = (path["method"], path["path"])
        if key not in seen:
            seen.add(key)
            deduplicated_paths.append(path)
    return deduplicated_paths


def _get_drf_paths(urlconfs: List[Optional[str]]) -> List[Dict[str, str]]:
    from rest_framework.schemas.generators import EndpointEnumerator

    enumerators = [EndpointEnumerator(urlconf=urlconf) for urlconf in urlconfs]
    return [
        {
            "method": method.upper(),
            "path": path,
        }
        for enumerator in enumerators
        for path, method, _ in enumerator.get_api_endpoints()
        if method not in ["HEAD", "OPTIONS"]
    ]


def _get_drf_callbacks(urlconfs: List[Optional[str]]) -> Set[Callable]:
    from rest_framework.schemas.generators import EndpointEnumerator

    enumerators = [EndpointEnumerator(urlconf=urlconf) for urlconf in urlconfs]
    return {callback for enumerator in enumerators for _, _, callback in enumerator.get_api_endpoints()}


def _get_drf_schema(urlconfs: List[Optional[str]]) -> Optional[Dict[str, Any]]:
    from rest_framework.schemas.openapi import SchemaGenerator

    schemas = []
    # AssertionError is raised if uritemplate or inflection are not installed (required for OpenAPI schema support)
    # AttributeError is raised if app is using CoreAPI schema (deprecated) instead of OpenAPI
    with contextlib.suppress(AssertionError, AttributeError):
        for urlconf in urlconfs:
            generator = SchemaGenerator(urlconf=urlconf)
            schema = generator.get_schema()
            if schema is not None and len(schema["paths"]) > 0:
                schemas.append(schema)
    return None if len(schemas) != 1 else schemas[0]  # type: ignore[return-value]


def _get_ninja_paths(urlconfs: List[Optional[str]]) -> List[Dict[str, str]]:
    endpoints = []
    for api in _get_ninja_api_instances(urlconfs=urlconfs):
        schema = api.get_openapi_schema()
        for path, operations in schema["paths"].items():
            for method, operation in operations.items():
                if method not in ["HEAD", "OPTIONS"]:
                    endpoints.append(
                        {
                            "method": method,
                            "path": path,
                            "summary": operation.get("summary"),
                            "description": operation.get("description"),
                        }
                    )
    return endpoints


def _get_ninja_callbacks(urlconfs: List[Optional[str]]) -> Set[Callable]:
    return {
        path_view.get_view()
        for api in _get_ninja_api_instances(urlconfs=urlconfs)
        for _, router in api._routers
        for path_view in router.path_operations.values()
    }


def _get_ninja_schema(urlconfs: List[Optional[str]]) -> Optional[Dict[str, Any]]:
    schemas = []
    for api in _get_ninja_api_instances(urlconfs=urlconfs):
        schema = api.get_openapi_schema()
        if len(schema["paths"]) > 0:
            schemas.append(schema)
    return None if len(schemas) != 1 else schemas[0]


def _get_ninja_api_instances(
    urlconfs: Optional[List[Optional[str]]] = None,
    patterns: Optional[List[Any]] = None,
) -> Set[NinjaAPI]:
    from ninja import NinjaAPI

    if urlconfs is None:
        urlconfs = [None]
    if patterns is None:
        patterns = []
        for urlconf in urlconfs:
            patterns.extend(get_resolver(urlconf).url_patterns)

    apis: Set[NinjaAPI] = set()
    for p in patterns:
        if isinstance(p, URLResolver):
            if p.app_name != "ninja":
                apis.update(_get_ninja_api_instances(patterns=p.url_patterns))
            else:
                for pattern in p.url_patterns:
                    if isinstance(pattern, URLPattern) and pattern.lookup_str.startswith("ninja."):
                        callback_keywords = getattr(pattern.callback, "keywords", {})
                        if isinstance(callback_keywords, dict):
                            api = callback_keywords.get("api")
                            if isinstance(api, NinjaAPI):
                                apis.add(api)
                                break
    return apis


def _get_django_paths(urlconfs: Optional[List[Optional[str]]] = None) -> List[Dict[str, str]]:
    if urlconfs is None:
        urlconfs = [None]
    return [
        {
            "method": method.upper(),
            "path": _transform_path(regex),
        }
        for urlconf in urlconfs
        for callback, regex, _, _ in extract_views_from_urlpatterns(get_resolver(urlconf).url_patterns)
        if hasattr(callback, "view_class") and issubclass(callback.view_class, View)
        for method in callback.view_class.http_method_names
        if method != "options" and hasattr(callback.view_class, method)
    ]


def _get_django_callbacks(urlconfs: Optional[List[Optional[str]]] = None) -> Set[Callable]:
    if urlconfs is None:
        urlconfs = [None]
    return {
        callback
        for urlconf in urlconfs
        for callback, _, _, _ in extract_views_from_urlpatterns(get_resolver(urlconf).url_patterns)
    }


def _transform_path(path: str) -> str:
    path = simplify_regex(path)
    return re.sub(r"<(?:(?P<converter>[^>:]+):)?(?P<parameter>\w+)>", r"{\g<parameter>}", path)


def _check_import(name: str) -> bool:
    try:
        import_module(name)
        return True
    except ImportError:
        return False



from __future__ import annotations

import asyncio
import sys
import threading
import traceback
from collections import Counter
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Set

from apitally.client.sentry import get_sentry_event_id_async


MAX_EXCEPTION_MSG_LENGTH = 2048
MAX_EXCEPTION_TRACEBACK_LENGTH = 65536


@dataclass(frozen=True)
class ServerError:
    consumer: Optional[str]
    method: str
    path: str
    type: str
    msg: str
    traceback: str


class ServerErrorCounter:
    def __init__(self) -> None:
        self.error_counts: Counter[ServerError] = Counter()
        self.sentry_event_ids: Dict[ServerError, str] = {}
        self._lock = threading.Lock()
        self._tasks: Set[asyncio.Task] = set()

    def add_server_error(self, consumer: Optional[str], method: str, path: str, exception: BaseException) -> None:
        if not isinstance(exception, BaseException):
            return  # pragma: no cover
        with self._lock:
            server_error = ServerError(
                consumer=consumer,
                method=method.upper(),
                path=path,
                type=get_exception_type(exception),
                msg=get_truncated_exception_msg(exception),
                traceback=get_truncated_exception_traceback(exception),
            )
            self.error_counts[server_error] += 1
        get_sentry_event_id_async(lambda event_id: self.sentry_event_ids.update({server_error: event_id}))

    def get_and_reset_server_errors(self) -> List[Dict[str, Any]]:
        data: List[Dict[str, Any]] = []
        with self._lock:
            for server_error, count in self.error_counts.items():
                data.append(
                    {
                        "consumer": server_error.consumer,
                        "method": server_error.method,
                        "path": server_error.path,
                        "type": server_error.type,
                        "msg": server_error.msg,
                        "traceback": server_error.traceback,
                        "sentry_event_id": self.sentry_event_ids.get(server_error),
                        "error_count": count,
                    }
                )
            self.error_counts.clear()
            self.sentry_event_ids.clear()
        return data


def get_exception_type(exception: BaseException) -> str:
    exception_type = type(exception)
    return f"{exception_type.__module__}.{exception_type.__qualname__}"


def get_truncated_exception_msg(exception: BaseException) -> str:
    msg = str(exception).strip()
    if len(msg) <= MAX_EXCEPTION_MSG_LENGTH:
        return msg
    suffix = "... (truncated)"
    cutoff = MAX_EXCEPTION_MSG_LENGTH - len(suffix)
    return msg[:cutoff] + suffix


def get_truncated_exception_traceback(exception: BaseException) -> str:
    prefix = "... (truncated) ...\n"
    cutoff = MAX_EXCEPTION_TRACEBACK_LENGTH - len(prefix)
    lines = []
    length = 0
    if sys.version_info >= (3, 10):
        traceback_lines = traceback.format_exception(exception)
    else:
        traceback_lines = traceback.format_exception(type(exception), exception, exception.__traceback__)
    for line in traceback_lines[::-1]:
        if length + len(line) > cutoff:
            lines.append(prefix)
            break
        lines.append(line)
        length += len(line)
    return "".join(lines[::-1]).strip()



from __future__ import annotations

import threading
from typing import Any, Dict, List, Optional, Set, Union


class Consumer:
    def __init__(self, identifier: str, name: Optional[str] = None, group: Optional[str] = None) -> None:
        self.identifier = str(identifier).strip()[:128]
        self.name = str(name).strip()[:64] if name else None
        self.group = str(group).strip()[:64] if group else None

    @classmethod
    def from_string_or_object(cls, consumer: Optional[Union[str, Consumer]]) -> Optional[Consumer]:
        if not consumer:
            return None
        if isinstance(consumer, Consumer):
            return consumer
        consumer = str(consumer).strip()
        if not consumer:
            return None
        return cls(identifier=consumer)

    def update(self, name: str | None = None, group: str | None = None) -> bool:
        name = str(name).strip()[:64] if name else None
        group = str(group).strip()[:64] if group else None
        updated = False
        if name and name != self.name:
            self.name = name
            updated = True
        if group and group != self.group:
            self.group = group
            updated = True
        return updated


class ConsumerRegistry:
    def __init__(self) -> None:
        self.consumers: Dict[str, Consumer] = {}
        self.updated: Set[str] = set()
        self._lock = threading.Lock()

    def add_or_update_consumer(self, consumer: Optional[Consumer]) -> None:
        if not consumer or (not consumer.name and not consumer.group):
            return  # Only register consumers with name or group set
        with self._lock:
            if consumer.identifier not in self.consumers:
                self.consumers[consumer.identifier] = consumer
                self.updated.add(consumer.identifier)
            elif self.consumers[consumer.identifier].update(name=consumer.name, group=consumer.group):
                self.updated.add(consumer.identifier)

    def get_and_reset_updated_consumers(self) -> List[Dict[str, Any]]:
        data: List[Dict[str, Any]] = []
        with self._lock:
            for identifier in self.updated:
                if consumer := self.consumers.get(identifier):
                    data.append(
                        {
                            "identifier": consumer.identifier,
                            "name": str(consumer.name)[:64] if consumer.name else None,
                            "group": str(consumer.group)[:64] if consumer.group else None,
                        }
                    )
            self.updated.clear()
        return data
